""" Analog ensemble variants.

Currently, :class:`~engine.neuron.analog.AnalogNeuron` or :class:`~engine.neuron.analog.oscillator.OscillatorNeuron`
can be used as base classes. Analog neurons simply add the input to the state stored in `voltage`, while oscillators
are capable of generating an arbitrary combination of sine waves, with basic support for phase-amplitude coupling
via component amplitude oscillations.

Ensembles can be initialized programmatically or from a configuration YAML containing neuron parameters and their
generation method (fixed, zipped, grid, or from distribution). Forward calls invoke the parent neuron class method.

See Also
--------
:class:`~engine.neuron.analog.AnalogNeuron`
:class:`~engine.neuron.analog.oscillator.OscillatorNeuron`

"""
import torch

from sapicore.engine.ensemble import Ensemble
from sapicore.engine.neuron.analog import AnalogNeuron
from sapicore.engine.neuron.analog.oscillator import OscillatorNeuron

__all__ = ("AnalogEnsemble", "OscillatorEnsemble")


class AnalogEnsemble(Ensemble, AnalogNeuron):
    """Ensemble of generic analog neurons."""

    def __init__(self, **kwargs):
        """Constructs an analog ensemble instance, inheriting attributes from
        :class:`~engine.neuron.analog.AnalogNeuron`.

        """
        # invoke parent constructor and initialization method(s).
        super().__init__(**kwargs)


class OscillatorEnsemble(Ensemble, OscillatorNeuron):
    """Ensemble of oscillators.

    Parameters
    ----------
    num_wave_comps: int
        Number of components in each composite sinusoidal.

    Note
    ----
    To initialize an ensemble where the number of frequency components differs across elements,
    set `num_wave_comps` to the maximal number and unnecessary elements to 0.

    For instance, frequencies: [[34.0, 41.0, 50.0], [30.0, 0.0, 0.0]] would yield a 2-ensemble where the first
    element is the sum of three sine components and the second element is a simple 30Hz wave.

    Sapicore does not currently support the configuration syntax [[34.0, 41.0, 50.0], [30.0]] due to the
    inconvenience of dealing with variable-length tensors.

    """

    _extensible_props_: tuple[str] = ("amplitudes", "frequencies", "phases", "amp_freq")

    def __init__(self, num_wave_comps: int = 1, **kwargs):
        """Constructs an oscillator ensemble instance, inheriting attributes from
        :class:`~engine.ensemble.Ensemble` and :class:`~engine.neuron.analog.oscillator.OscillatorNeuron`."""
        # generic ensemble initialization with configurable and loggable tensors.
        super().__init__(**kwargs)

        # number of frequency components each oscillator consists of (use max if variable).
        self.num_wave_comps = num_wave_comps

        # expand extensible tensors to accommodate 2D specification format of multi-component sine waves.
        for prop in self.extensible_props:
            temp = getattr(self, prop)
            zeros = torch.zeros(size=(self.num_units, self.num_wave_comps), dtype=torch.float, device=self.device)

            setattr(self, prop, zeros + temp.reshape(self.num_units, 1))

        # create the wave iterator for each element of this oscillator ensemble.
        self.register_waveforms()

    def heterogenize(self, num_combinations: int, unravel: bool = True):
        """Ensures that calling an oscillator's `heterogenize()` will result in a correct row-wise treatment of
        tensor attributes and update the waveform iterators.

        Note
        ----
        This is necessary, as `heterogenize` is sometimes invoked from ambiguous ensemble classes (e.g., in `Network`,
        when ensembles are initialized from file and dynamically imported).

        """
        super().heterogenize(num_combinations=self.num_units, unravel=False)
        self.register_waveforms()



""" Simple simulation pipeline. """
import logging
import os

from argparse import ArgumentParser
from datetime import datetime

import matplotlib

import torch
from torch import Tensor

from sapicore.data import Data
from sapicore.model import Model
from sapicore.pipeline import Pipeline
from sapicore.engine.network import Network

from sapicore.utils.constants import DT, SEED, TIME_FORMAT, PLOT_BACKEND
from sapicore.utils.seed import fix_random_seed
from sapicore.utils.io import ensure_dir, log_settings, save_yaml, plot_tensorboard

__all__ = ("SimpleSimulator",)


class SimpleSimulator(Pipeline):
    """A prototypical simulation pipeline, bringing together data loading, network construction,
    model specification, disk I/O, and visualization.

    This pipeline can be configured by providing a dictionary or a path to a YAML.

    Parameters
    ----------
    configuration: dict or str
        Pipeline script configuration dictionary or path to YAML.

    data: Data or Tensor, optional
        Sapicore :class:`~data.Data` object or Tensor to use during model training.

    Note
    ----
    For a more complex demonstration involving data processing, selection, sampling, and cross-validation,
    see `tutorials/basic_experiment.py`. If using this script with data (be it Data or Tensor),
    it should be curated and coerced to the shape buffer X labels ahead of time.

    """

    def __init__(self, configuration: dict | str, data: Data | Tensor = None, **kwargs):
        super().__init__(config_or_path=configuration)
        matplotlib.use(PLOT_BACKEND)

        # set hardware device from configuration, default to GPU if available or CPU otherwise.
        self.device = self.configuration.get("device", "cpu" if not torch.cuda.is_available() else "cuda:0")

        # reference to data object containing buffer to be streamed to the network.
        self.data = data

        # project root directory, defaults to configuration YAML directory if given or empty otherwise.
        self.project_root = os.path.join(configuration, "..") if isinstance(configuration, str) else ""

        # prepare the environment by validating paths and setting a random seed for reproducibility.
        self.initialize(seed=kwargs.get("seed"))

    def initialize(self, seed: int = None):
        """Initializes the pipeline. Validates data object and project root path given in the configuration,
        sets random number generation seed, and applies log message configuration.

        Parameters
        ----------
        seed: int, optional
            Random seed to be used throughout the simulation. Applied to all relevant libraries.
            If not provided, looks up "seed" in the configuration dictionary.
            If not found, the project default :attr:`utils.constants.SEED` is used to ensure reproducibility.

        Warning
        -------
        Deterministic behavior is vital when conducting reproducible experiments. This is one of many safeguards
        included in Sapicore. Users who wish to modify this default behavior will have to consciously override
        :meth:`~SimpleSimulator.initialize`.

        """
        # update project root path from YAML configuration file if given.
        yaml_root = self.configuration.get("root")
        if yaml_root:
            if os.path.isabs(yaml_root):
                # if the given path is absolute, take it as is.
                self.project_root = yaml_root
            else:
                # otherwise, append it to the current value in project_root.
                # if the latter is empty, results in yaml_root relative to current working directory.
                self.project_root = os.path.join(self.project_root, yaml_root)

        # update full project root in configuration dictionary (used during network instantiation).
        self.project_root = os.path.realpath(self.project_root)
        self.configuration["root"] = self.project_root

        # prepare the dataset object by calling it (derived data classes implement __call__ appropriately).
        if self.data:
            if not (isinstance(self.data, Tensor) or isinstance(self.data, Data)):
                raise TypeError(f"Data object provided is of the unsupported type {type(self.data)}")

            if isinstance(self.data, Data):
                self.data = self.data.load()

            # moves data to the correct device (syntax is supported as it inherits from torch's Dataset).
            self.data = self.data[:].to(self.device)

        # set random seed for entire simulation based on config, kwarg, or default (in this order of preference).
        random_seed = seed if seed is not None else self.configuration.get("seed", SEED)
        fix_random_seed(random_seed)

        # apply default logger configuration.
        log_settings()

    def run(self):
        """Runs the generic simulator pipeline based on the given configuration."""
        # create timestamped directory for this simulation run in the project root.
        stamp = datetime.now().strftime(TIME_FORMAT)
        run_dir = ensure_dir(os.path.join(self.project_root, stamp))

        # load simulation settings from configuration dictionary.
        steps = self.configuration.get("simulation", {}).get("steps")
        disk = self.configuration.get("simulation", {}).get("disk")
        tensorboard = self.configuration.get("simulation", {}).get("tensorboard")

        # create data and tensorboard subdirectories.
        tb_dir = ensure_dir(os.path.join(run_dir, "tb"))
        data_dir = ensure_dir(os.path.join(run_dir, "data"))

        # save a copy of the master configuration file to this run's directory for future reference.
        save_yaml(self.configuration, os.path.join(run_dir, self.configuration.get("identifier", "config") + ".yaml"))

        # initialize a network using the given configuration and attach it to a model instance.
        logging.info("Instantiating the network.")
        model = Model(network=Network(configuration=self.configuration, device=self.device))

        # describe the network to the user.
        logging.info(model.network)

        if disk:
            # register data accumulator hooks for streaming intermediate output to disk (automatically applied).
            model.network.add_data_hook(data_dir, steps)

            # save an SVG plot of the network architecture.
            model.network.draw(path=run_dir)

        if not self.data:
            steps = self.configuration.get("simulation", {}).get("steps", {})
            if not steps:
                raise RuntimeError(
                    "Cannot run a simulation when neither data nor a number of simulation steps"
                    "(under the key simulation/steps in the configuration file) was given."
                )

            # if data not provided, set data to 0s of shape duration X root_node_num_units
            self.data = torch.zeros((steps, model.network[model.network.roots[0]].num_units), device=self.device)

            # if constant current value provided, add it to the dummy data tensor (useful for rudimentary testing).
            self.data = self.data + self.configuration.get("simulation", {}).get("current", 0.0)

        # fit the model by passing `data` buffer to the network (`duration` controls exposure time to each sample).
        logging.info(f"Simulating {steps} steps at a resolution of {DT} ms.")
        model.fit(data=self.data, duration=self.configuration.get("duration", 1))

        # optional tensorboard logging.
        if tensorboard:
            plot_tensorboard(self.configuration, data_dir, tb_dir)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "-config",
        action="store",
        dest="config",
        metavar="FILE",
        required=True,
        help="Path to simulator pipeline configuration file.",
    )
    args = parser.parse_args()

    # run the simulation with the given configuration file.
    SimpleSimulator(configuration=args.config).run()



from typing import Callable
from datetime import datetime as dt

import pytest
import torch

from sapicore.engine.ensemble.analog import AnalogEnsemble
from sapicore.engine.ensemble.spiking import LIFEnsemble, IZEnsemble
from sapicore.utils.integration import RungeKutta

import matplotlib.pyplot as plt

UNITS = 1
STEPS = 500
INTERVAL = 25
PLOT = False


class Curve:
    """Represents an ODE with an analytical solution, sets an initial condition `t0`,
    and computes its value `y0` at that point.

    Calling the object with an integer `t` will evaluate the curve at `t`.

    """

    def __init__(self, curve: Callable, ode: Callable, t0: float):
        self.curve = curve
        self.ode = ode
        self.t0 = t0
        self.y0 = self.curve(t=self.t0)

    def __call__(self, t: int):
        return self.curve(t=t)


class TestIntegration:
    @pytest.mark.functional
    def test_numeric_correctness(self):
        """Demonstrates correctness of Euler and RK4 using example ODEs with simple analytical solutions."""
        # define analytic curves, their respective ODEs, and t0 at which to start the simulation.
        curves = [
            Curve(curve=lambda t: torch.e**t, ode=lambda x: x, t0=1.0),
            Curve(curve=lambda t: (t**2.0) / 4.0, ode=lambda x: torch.sqrt(x), t0=4.0),
        ]

        for curve in curves:
            # initialize two analog ensembles that integrate using Euler/RK4.
            analog_euler = AnalogEnsemble(num_units=UNITS, integrator=RungeKutta(order=1), equation=curve.ode)
            analog_rk4 = AnalogEnsemble(num_units=UNITS, integrator=RungeKutta(order=4), equation=curve.ode)

            # update initial voltages to the value of the curve at its t0.
            analog_euler.voltage = torch.tensor(curve.y0)
            analog_rk4.voltage = torch.tensor(curve.y0)

            # compute the underlying function and Euler/RK4 approximations in the interval [0, INTERVAL).
            underlying = []
            euler = []
            rk4 = []

            for j in range(INTERVAL):
                # underlying curve at this time point.
                underlying.append(curve(t=curve.t0 + (j + 1.0) * analog_euler.dt))

                # integrate ensembles directly, bypassing forward().
                analog_euler.voltage = analog_euler.integrate()
                analog_rk4.voltage = analog_rk4.integrate()

                # store voltage values for the two integration modes.
                euler.append(analog_euler.voltage.mean().item())
                rk4.append(analog_rk4.voltage.mean().item())

            # plots show RK4 outperforming Euler for our test cases y(t) = e^t and y(t) = t^2 / 4.
            if PLOT:
                plt.plot(underlying)
                plt.plot(rk4)
                plt.plot(euler)
                plt.title("Ground Truth (Blue), RK4 (Orange), Euler (Green)")

                plt.show()

    @pytest.mark.parametrize("ref", [LIFEnsemble, IZEnsemble], ids=["LIF", "IZ"])
    @pytest.mark.functional
    def test_spiking_comparison(self, ref: type):
        """Shows that Euler/RK4 integration makes a difference for the default spiking neuron models (IZ, LIF)."""
        # initialize two spiking ensembles with Euler/RK4 integrators.
        euler = ref(identifier="Euler", num_units=UNITS, integrator=RungeKutta(order=1))
        rk4 = ref(identifier="RK4", num_units=UNITS, integrator=RungeKutta(order=4))

        # input current values were selected to trigger firing given the default LIF and IZ parameters.
        data = torch.zeros(UNITS) + (105.0 if ref is LIFEnsemble else 10.0)

        euler_values = torch.zeros((STEPS, UNITS))
        rk4_values = torch.zeros((STEPS, UNITS))

        start = dt.now()
        for i in range(STEPS):
            euler_values[i, :] = euler(data)["voltage"].cpu()
        euler_duration = dt.now() - start

        start = dt.now()
        for i in range(STEPS):
            rk4_values[i, :] = rk4(data)["voltage"].cpu()
        rk4_duration = dt.now() - start

        md = (euler_values - rk4_values).mean()
        sd = (euler_values - rk4_values).std()

        if PLOT:
            print(
                f"\nCompared to Euler, RK4 took x{rk4_duration / euler_duration}. "
                f"At DT={euler.dt}, Steps={STEPS}, Mean Euler-RK4 difference: {md:.2f} (SD={sd:.2f})"
            )

            plt.plot(euler_values.tolist())
            plt.plot(rk4_values.tolist())
            plt.title("Euler/RK4 Voltage Traces")
            plt.show()


if __name__ == "__main__":
    pytest.main(args=["-s", "-v"])



""" Spiking ensemble variants.

Sapicore provides leaky integrate-and-fire (LIF) and Izhikevich neuron implementations.

See Also
--------
:class:`~engine.neuron.spiking.LIF.LIFNeuron`
:class:`~engine.neuron.spiking.IZ.IZNeuron`

"""
from sapicore.engine.ensemble import Ensemble

from sapicore.engine.neuron.spiking.IZ import IZNeuron
from sapicore.engine.neuron.spiking.LIF import LIFNeuron

__all__ = ("LIFEnsemble", "IZEnsemble")


class LIFEnsemble(Ensemble, LIFNeuron):
    """Ensemble of LIF neurons, inheriting its attributes from :class:`~engine.neuron.spiking.LIF.LIFNeuron`."""

    def __init__(self, **kwargs):
        """Constructs a LIF ensemble, inheriting its attributes from :class:`~engine.neuron.spiking.LIF.LIFNeuron`."""
        super().__init__(**kwargs)

    def heterogenize(self, num_combinations: int, unravel: bool = True):
        """Ensures that calling :meth:`~heterogenize` will also update the voltage to what could be
        a changed resting potential."""

        super().heterogenize(num_combinations=self.num_units, unravel=unravel)
        self.voltage = self.volt_rest.detach().clone()


class IZEnsemble(Ensemble, IZNeuron):
    """Ensemble of IZ neurons, inheriting its attributes from :class:`~engine.neuron.spiking.IZ.IZNeuron`."""

    def __init__(self, **kwargs):
        """Constructs a LIF ensemble, inheriting its attributes from :class:`~engine.neuron.spiking.LIF.LIFNeuron`."""
        super().__init__(**kwargs)

    def heterogenize(self, num_combinations: int, unravel: bool = True):
        """Ensures that calling :meth:`~heterogenize` will also update the voltage to what could be
        a changed resting potential."""

        super().heterogenize(num_combinations=self.num_units, unravel=unravel)
        self.voltage = self.c.detach().clone()



""" Leaky Integrate-and-Fire neuron model (LIF). """
import torch

from torch import Tensor
from torch import as_tensor
from torch.nn.functional import relu

from sapicore.engine.neuron.spiking import SpikingNeuron


class LIFNeuron(SpikingNeuron):
    """Base class for the leaky integrate-and-fire neuron model (LIF).

    Extends :class:`~engine.neuron.spiking.SpikingNeuron` by adding necessary parameters and implementing
    the LIF :meth:`~forward` method efficiently for 1D tensors of arbitrary size.
    User may implement subclasses to differentiate variants along additional dimensions.

    LIF units own the following attributes in addition to those inherited from
    :class:`~engine.neuron.spiking.SpikingNeuron`:

    Parameters
    ----------
    volt_thresh: float or Tensor
        Spike threshold (defaults to -55.0).

    volt_rest: float or Tensor
        Resting potential (defaults to -75.0).

    leak_gl: float or Tensor
        Leak conductance (defaults to 5.0).

    tau_mem: float or Tensor
        Membrane time constant (e.g., 5.0).

    tau_ref: float or Tensor
        Refractory period (e.g., 1.0).

    cycle_length: int, optional
        Oscillatory cycle period, required to time optional resetting of the refractory period.

    release_phase: int, optional
        Oscillation phase at which to release all neurons from refractory mode, if required.

    References
    ----------
    `LIF Tutorial <https://compneuro.neuromatch.io/tutorials/W2D3_BiologicalNeuronModels/student/W2D3_Tutorial1.html>`_
        :meth:`~engine.neuron.spiking.LIF.LIFNeuron.forward` implements the algorithm described in this tutorial.

    """

    # loggable properties are identical to those of generic spiking neurons, so no overriding is necessary.
    _config_props_: tuple[str] = (
        "volt_thresh",
        "volt_rest",
        "leak_gl",
        "tau_mem",
        "tau_ref",
        "cycle_length",
        "release_phase",
    )

    def __init__(
        self,
        volt_thresh=-55.0,
        volt_rest=-75.0,
        leak_gl=5.0,
        tau_mem=5.0,
        tau_ref=1.0,
        cycle_length: int = None,
        release_phase: int = None,
        **kwargs
    ):
        """Instantiates a single LIF neuron with its default configurable and simulation attributes."""
        # register universal attributes inherited from `SpikingNeuron`, `Neuron`, and `Component`.
        super().__init__(**kwargs)

        # configurable attributes specific to LIF neurons, over and above `SpikingNeuron`.
        self.volt_thresh = torch.zeros(1, dtype=torch.float, device=self.device) + as_tensor(
            volt_thresh, device=self.device
        )
        self.volt_rest = torch.zeros(1, dtype=torch.float, device=self.device) + as_tensor(
            volt_rest, device=self.device
        )
        self.leak_gl = torch.zeros(1, dtype=torch.float, device=self.device) + as_tensor(leak_gl, device=self.device)
        self.tau_mem = torch.zeros(1, dtype=torch.float, device=self.device) + as_tensor(tau_mem, device=self.device)
        self.tau_ref = torch.zeros(1, dtype=torch.float, device=self.device) + as_tensor(tau_ref, device=self.device)

        # optional periodic LIF features.
        if cycle_length is not None:
            self.cycle_length = torch.zeros(1, dtype=torch.int, device=self.device) + cycle_length
        else:
            self.cycle_length = None

        if release_phase is not None:
            self.release_phase = torch.zeros(1, dtype=torch.int, device=self.device) + release_phase
        else:
            self.release_phase = None

        # simulation/state attributes specific to LIF, over and above `SpikingNeuron`.
        self.input = torch.zeros_like(self.volt_rest)
        self.spiked = torch.zeros_like(self.volt_rest)
        self.voltage = self.volt_rest.detach().clone()  # initialize membrane potential to resting potential.
        self.refractory_steps = torch.zeros_like(self.volt_rest).int()

        # difference equation of this neuron model, to be used with numeric approximation methods.
        self.equation = lambda x, data: (-(x - self.volt_rest) + (data / self.leak_gl)) / self.tau_mem

    def forward(self, data: Tensor) -> dict:
        """LIF forward method, to be used by ensemble or singleton 1D tensors.

        Parameters
        ----------
        data: Tensor
            External input to be processed by this LIF unit.

        Returns
        -------
        dict
            Dictionary containing the numeric state tensors `voltage`, `spiked`, and `input`.

        """
        if self.gate_:
            # update input data based on gating signal.
            data = self.gate * data
            self.gate_ = False

        if self.teach_:
            # teaching signals override gating signals.
            data = self.teach
            self.teach_ = False

        # update internal representation of input current for tensorboard logging purposes.
        self.input = as_tensor(data, device=self.device)

        # detect spikes in a pytorch-friendly way by thresholding the voltage attribute.
        spiked_prev = self.voltage >= self.volt_thresh

        # reset neurons that spiked in the previous iteration to their resting potential, or add zero otherwise.
        self.voltage = self.voltage + spiked_prev.int() * (self.volt_rest - self.voltage)
        self.spiked = spiked_prev.int()

        # add refractory steps for the neurons that spiked, otherwise do nothing.
        self.refractory_steps = self.refractory_steps + spiked_prev * torch.div(self.tau_ref, self.dt).int()

        # integrate input current, now that we know who spiked and updated their refractory status.
        self.voltage = self.integrate(data=data)

        # clamp membrane potential to the threshold `volt_thresh` if necessary.
        threshold_exceeded = (self.voltage >= self.volt_thresh).int()
        self.voltage = self.voltage + threshold_exceeded * (self.volt_thresh - self.voltage)

        # figure out which neurons in the tensor are currently refractory.
        refractory = self.refractory_steps > 0
        over_resting = self.voltage > self.volt_rest

        # if refractory, add difference between resting and current membrane potential. Otherwise, do nothing.
        self.voltage = self.voltage + (refractory.int() * over_resting.int()) * (self.volt_rest - self.voltage)

        # decrement refractory steps, taking care to zero out negatives.
        self.refractory_steps = relu(self.refractory_steps - 1)
        self.simulation_step += 1

        if self.release_phase is not None and self.cycle_length is not None:
            # voltage will start to accumulate at a particular phase, canceling the refractory period across units.
            reset = torch.full_like(self.cycle_length, self.simulation_step) % self.cycle_length == self.release_phase

            self.refractory_steps = self.refractory_steps * (~reset) + torch.zeros_like(self.refractory_steps) * reset
            self.voltage = self.voltage * (~reset) + self.volt_rest * reset

        # return current state(s) of loggable attributes as a dictionary.
        return self.loggable_state()



"""Train and test neuromorphic models.

Models are networks endowed with the methods :meth:`fit`, :meth:`predict`, :meth:`similarity`, :meth:`load`,
and :meth:`save`. This class provides an ML-focused API for the training and usage of sapicore
:class:`~engine.network.Network` output for practical purposes.

"""
from typing import Callable, Sequence

import dill
import os

from alive_progress import alive_bar

import torch
from torch import Tensor

from sapicore.engine.network import Network
from sapicore.utils.io import ensure_dir


class Model:
    """Model base class.

    Loosely follows the design of scikit-learn's :class:`sklearn.base.BaseEstimator` interface.

    Note
    ----
    In a machine learning context, spiking networks can be used in diverse ways, e.g. as classifiers
    or as generative models. While :mod:`engine` is mostly about form (network architecture and information flow),
    :mod:`model` is all about function (how to fit the model to data and how to utilize it once trained).

    """

    def __init__(self, network: Network = None, **kwargs):
        # store a reference to one network object.
        self.network = network

    def serve(
        self, data: Tensor | Sequence[Tensor], duration: int | Sequence[int], rinse: int | Sequence[int] = 0, **kwargs
    ):
        """Applies :meth:`engine.network.Network.forward` sequentially on a batch of buffer `data`.

        Parameters
        ----------
        data: Tensor or Sequence of Tensor
            2D tensor(s) of data buffer to be fed to the root ensemble(s) of this object's `network`,
            formatted sample X feature.

        duration: int or Sequence of int
            Duration of sample presentation. Simulates duration of exposure to a particular input.
            If a list or a tensor is provided, the i-th sample in the batch is maintained for `duration[i]` steps.

        rinse: int or Sequence of int
            Null stimulation steps (0s in-between samples).
            If a list or a tensor is provided, the i-th sample is followed by `rinse[i]` rinse steps.
        Each sample `i` is presented for `duration[i]`, followed by all 0s stimulation for `rinse[i]`.

        """
        # wrap 2D tensor data in a list if need be, to make subsequent single- and multi-root operations uniform.
        if not isinstance(data, Sequence):
            data = [data]

        num_samples = data[0].shape[0]

        with alive_bar(total=num_samples, force_tty=True) as bar:
            for i in range(num_samples):
                # compute presentation and rinse durations.
                stim_steps = duration if isinstance(duration, int) else duration[i]
                rinse_steps = rinse if isinstance(rinse, int) else rinse[i]

                for k in range(stim_steps):
                    # stimulus presentation.
                    self.network.forward([data[j][i, :] for j in range(len(data))])

                for _ in range(rinse_steps):
                    # inter-stimulus interval.
                    self.network.forward([torch.zeros_like(data[j][i, :]) for j in range(len(data))])

                # advance progress bar.
                bar()

    def fit(
        self, data: Tensor | Sequence[Tensor], duration: int | Sequence[int], rinse: int | Sequence[int] = 0, **kwargs
    ):
        """Serves a batch of data, then turns off learning for all synapses.

        Warning
        -------
        :meth:`~model.Model.fit` does not return intermediate output. Users should register forward hooks to
        efficiently stream data to memory or disk throughout the simulation
        (see :meth:`~engine.network.Network.add_data_hook`).

        """
        self.serve(data, rinse, duration)

        # after fitting a model, learning is turned off for all synapses by default.
        for synapse in self.network.get_synapses():
            synapse.set_learning(False)

    def predict(self, data: Tensor, labels: Sequence, **kwargs) -> Sequence:
        """Predicts the labels of `data`.

        Parameters
        ----------
        data: Tensor
            Standalone 2D tensor of data buffer, sample X feature.

        labels: Sequence
            Label values corresponding to classification layer cell indices.

        Returns
        -------
        Sequence
            Vector of predicted labels.

        """
        raise NotImplementedError

    def similarity(self, data: Tensor, metric: str | Callable, **kwargs) -> Tensor:
        """Performs a similarity analysis on network responses to `data`, yielding a pairwise distance matrix.

        Parameters
        ----------
        data: Tensor
            2D tensor of data buffer, formatted sample X feature.

        metric: str or Callable
            Distance metric to be used. If a string value is provided, it should correspond to one of the available
            :mod:`scipy.spatial.distance` metrics. If a custom function is provided, it should accept the
            `data` tensor and return a scalar corresponding to their distance.

        """
        raise NotImplementedError

    def save(self, path: str):
        """Saves the :class:`engine.network.Network` object owned by this model to `path`.

        Parameters
        ----------
        path: str
            Destination path, inclusive of the file to which the network should be saved.

        Note
        ----
        The default implementation uses :meth:`torch.save`, for the common case where files are used.
        Users may override this method when other formats are called for.

        """
        ensure_dir(os.path.dirname(path))
        torch.save(self.network, path, pickle_module=dill)

    def load(self, path: str) -> Network:
        """Loads a :class:`engine.network.Network` from `path` and assigns it to this object's `network` attribute.

        Parameters
        ----------
        path: str
            Path to the file containing the model.

        Returns
        -------
        Network
            A reference to the loaded :class:`engine.network.Network` object in case it is
            required by the calling function.

        Note
        ----
        The default implementation uses :meth:`torch.load`, for the common case where files are used.
        Users may override this method when other formats are called for.

        """
        if os.path.exists(path):
            self.network = torch.load(path, pickle_module=dill)

        return self.network



""" Random number generation and reproducibility. """
import random
import numpy as np

import torch
from torch.backends import cudnn


def fix_random_seed(seed: int = None) -> int:
    """Controls random number generation reproducibility by setting a consistent seed throughout
    the project and forcing deterministic implementation of algorithms using the CUDNN backend.

    Parameters
    ----------
    seed: int, optional
        RNG seed value between 0 and 2**32 - 1. If `None`, a random value will be picked.

    Returns
    -------
    seed: int
        Returns the value of the seed in case a random value was picked and needs to be recorded.

    Examples
    --------
    Set random seed across libraries:

        >>> fix_random_seed(314)

    """
    if seed is None:
        seed = np.random.choice(2**32 - 1)

    if seed < 0 or seed > 2**32 - 1:
        raise ValueError(f"Seed needs to be a positive number between 0 and {2**32 - 1}.")

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.cuda.manual_seed(seed)
    cudnn.benchmark = False
    cudnn.deterministic = True

    return seed



""" Neuron unit tests. """
import os
import pytest

import torch

from sapicore.engine.neuron import Neuron
from sapicore.engine.neuron.analog import AnalogNeuron
from sapicore.engine.neuron.analog.oscillator import OscillatorNeuron
from sapicore.engine.neuron.spiking import SpikingNeuron
from sapicore.engine.neuron.spiking.LIF import LIFNeuron
from sapicore.engine.neuron.spiking.IZ import IZNeuron

from sapicore.tests import ROOT, TEST_DEVICE

STEPS = 100  # default number of simulation forward steps to execute during testing.
TEST_ROOT = os.path.join(ROOT, "tests", "neuron", "test_neuron")  # test root directory.


class TestNeuron:
    @pytest.mark.parametrize(
        "arg_",
        [
            Neuron(identifier="Test", device=TEST_DEVICE),
            SpikingNeuron(device=TEST_DEVICE),
        ],
        ids=["NEURON", "SPIKING"],
    )
    @pytest.mark.unit
    def test_base(self, arg_: Neuron):
        """Basic initialization test cases."""
        # verify that all attributes are initialized.
        assert all(
            hasattr(arg_, attr)
            for attr in (["identifier", "input", "voltage"] + ["spiked"] if isinstance(arg_, SpikingNeuron) else [])
        )

        # verify that loggable attributes are torch tensors on the correct device.
        assert all(type(getattr(arg_, attr)) is torch.Tensor for attr in arg_._loggable_props_)
        assert all(getattr(arg_, attr).device == torch.device(TEST_DEVICE) for attr in arg_._loggable_props_)

        # for coverage completeness.
        with pytest.raises(NotImplementedError):
            arg_.forward(data=torch.zeros(1))

    @pytest.mark.parametrize("arg_", [AnalogNeuron(device=TEST_DEVICE)], ids=["ANALOG"])
    @pytest.mark.unit
    def test_analog(self, arg_: AnalogNeuron):
        """Task initialization test cases."""
        # verify that all attributes are initialized.
        assert all(hasattr(arg_, attr) for attr in ["identifier", "input", "voltage"])

        # verify that loggable attributes are torch tensors on the correct device.
        assert all(type(getattr(arg_, attr)) is torch.Tensor for attr in arg_._loggable_props_)
        assert all(getattr(arg_, attr).device == torch.device(TEST_DEVICE) for attr in arg_._loggable_props_)

        # verify behavior of trivial forward method.
        init_voltage = arg_.voltage

        # adding a compatible shape data should update `voltage` buffer values.
        add_same_shape = arg_.forward(data=torch.ones_like(init_voltage, device=TEST_DEVICE))

        # adding a data tensor of a different shape should overwrite content of `voltage` buffer.
        add_diff_len = arg_.forward(
            data=torch.ones((init_voltage.shape[0] + 2, init_voltage.shape[0] + 2), device=TEST_DEVICE)
        )
        assert torch.all(add_diff_len["voltage"].eq(torch.ones_like(add_same_shape["voltage"])))
        assert torch.all(add_diff_len["voltage"].eq(arg_.voltage))

    @pytest.mark.parametrize(
        "arg_",
        [
            OscillatorNeuron(device=TEST_DEVICE),
            OscillatorNeuron(amplitudes=[5.0, 10.0], frequencies=[20.0, 40.0], device=TEST_DEVICE),
            OscillatorNeuron(amplitudes=[5.0, 10.0], frequencies=[20.0, 40.0], phases=[0.0, 0.0], device=TEST_DEVICE),
            OscillatorNeuron(
                device=TEST_DEVICE,
                amplitudes=[5.0, 10.0],
                frequencies=[20.0, 40.0],
                phases=[0.0, 0.0],
                amp_freq=[0.0, 20.0],
            ),
        ],
        ids=["DEFAULT", "AF", "AFP", "PAC"],
    )
    @pytest.mark.unit
    def test_oscillator(self, arg_: OscillatorNeuron):
        # verify that all attributes are initialized.
        assert all(hasattr(arg_, attr) for attr in ["identifier", "input", "voltage"])

        # SUGGEST add signal correctness verification of some sort.
        [arg_.forward(data=torch.ones_like(arg_.voltage, device=TEST_DEVICE)) for _ in range(STEPS)]

    @pytest.mark.parametrize(
        "arg_",
        [
            LIFNeuron(device=TEST_DEVICE),
            LIFNeuron(
                volt_thresh=-50.0,
                volt_rest=-60.0,
                leak_gl=5.0,
                tau_mem=5.0,
                tau_ref=2.0,
                device=TEST_DEVICE,
            ),
        ],
        ids=["DEFAULT", "CUSTOM"],
    )
    @pytest.mark.unit
    def test_lif(self, arg_: LIFNeuron):
        # verify that all attributes are initialized.
        assert all(hasattr(arg_, attr) for attr in ["identifier", "input", "voltage", "spiked"])

        # test forward pass
        [arg_.forward(data=torch.ones_like(arg_.voltage, device=TEST_DEVICE) * 100.0) for _ in range(STEPS)]

    @pytest.mark.parametrize(
        "arg_",
        [IZNeuron(device=TEST_DEVICE), IZNeuron(a=0.2, b=0.25, c=-60.0, d=4.0, device=TEST_DEVICE)],
        ids=["DEFAULT", "CUSTOM"],
    )
    @pytest.mark.unit
    def test_iz(self, arg_: IZNeuron):
        # verify that all attributes are initialized.
        assert all(hasattr(arg_, attr) for attr in ["identifier", "input", "voltage", "spiked"])

        # test forward pass
        [arg_.forward(data=torch.ones_like(arg_.voltage, device=TEST_DEVICE) * 10.0) for _ in range(STEPS)]


if __name__ == "__main__":
    pytest.main(args=["-s", "-v"])



""" Run this script or `pytest -s -v` from the sapicore source root directory to execute all test configurations.
Select or exclude marked tests by category with `-m` (unit, integration, functional, slow).

This script generates an HTML coverage report under tests/coverage_report.
"""
import os
import pytest

from coverage import Coverage
from argparse import ArgumentParser

from sapicore.utils.io import ensure_dir
from sapicore.tests import ROOT


def run_tests(root=ROOT):
    # parse runtime arguments.
    parser = ArgumentParser()
    parser.add_argument(
        "-m",
        action="store",
        dest="m",
        nargs="+",
        default="all",
        help="Test categories to run (unit, integration, functional, slow, or all).",
    )

    parser.add_argument("-v", action="store_true", dest="v", default=False, help="Verbose pytest output.")
    parser.add_argument("-s", action="store_true", dest="s", default=False, help="Show runtime output, no capture.")

    # parse arguments to pass along to pytest, e.g. test category selection.
    args = parser.parse_args()

    args_list = []
    if args.v:
        args_list += ["-v"]

    if args.s:
        args_list += ["-s"]

    if args.m and "all" not in args.m:
        args_list += ["-m " + " ".join(args.m)]

    # unavoidable--pytest refuses to work with .ini at root or accept absolute paths when executed from here.
    os.chdir(root)

    # initialize coverage.py object.
    cov = Coverage(omit=[os.path.join(root, folder, "*") for folder in ["scripts"]])
    cov.start()

    # run configurable tests.
    pytest.main(args=args_list)

    # generate HTML coverage report.
    cov.stop()
    cov.html_report(directory=ensure_dir(os.path.join(root, "tests", "scripts", "coverage_report")))


if __name__ == "__main__":
    run_tests()



""" Data sampling utility classes.

Leverages scikit-learn and pandas to implement versatile cross validation and sampling schemes.

"""
from pandas import DataFrame
from sklearn.model_selection import BaseCrossValidator

__all__ = ("BalancedSampler", "CV")


class BalancedSampler:
    """Sampler base class.

    Leverages pandas' group-by operation to parsimoniously perform (dis)proportionate random sampling,
    with or without replacement.

    Parameters
    ----------
    stratified: bool
        By default (`False`), attempts to draw the same number of buffer from each group regardless of imbalances.
        If `True`, uses stratified sampling (proportionate, preserving class imbalances).

    replace: bool
        Whether to sample with replacement. Defaults to `False`.

    Note
    ----
    :meth:`~data.Data.sample` accepts initialized objects of this class.
    Additional samplers may be implemented in future versions.

    """

    def __init__(self, stratified: bool = False, replace: bool = False):
        self.stratified = stratified
        self.replace = replace

    def __call__(self, frame: DataFrame, group_keys: str | list[str], n: int | float):
        """Draws buffer from `frame` after grouping it by `columns`.

        Parameters
        ----------
        frame: DataFrame
            Pandas dataframe containing descriptor labels.

        group_keys: str or list of str
            Names of descriptors (column headers) to group the dataframe by.

        n: int or float
            Specifies the number or fraction of buffer to draw from each group.
            While the former is more convenient for equal (disproportionate) sampling and the latter
            for stratified (proportionate) sampling, integer or float values can be used with either one.

        """
        # keep track of global indices for subset selection.
        frame["index"] = frame.index
        grouped = frame.groupby(group_keys, group_keys=False)

        # convert `n` to integer if need be.
        if isinstance(n, float):
            n = int(n * len(frame["index"].tolist()))

        subset = grouped.apply(lambda x: x.sample(n, replace=self.replace), include_groups=False)

        return subset["index"].tolist()


class CV:
    """Cross validation base class.

    Parameter
    ---------
    data: Data
        The :class:`data.Data` object for which to generate the cross validator.

    cross_validator: BaseCrossValidator
        A scikit-learn :class:`BaseCrossValidator` object.

    label_keys: str or list of str, optional
        One or more descriptor label names. Required for some CV schemes (e.g., for StratifiedKFold, it would be
        the descriptor whose class label counts should be consistent with their proportions in the full sample).

    group_key: str, optional
        A grouping descriptor key. Required for some CV schemes (e.g., GroupKFold, for generating batch-like
        folds with non-overlapping groups w.r.t. some descriptor).

    Note
    ----
    Some external libraries extend the cross validator offerings of scikit-learn.
    This implementation is compatible with any class derived from BaseCrossValidator.

    """

    def __init__(
        self,
        data,
        cross_validator: BaseCrossValidator,
        label_keys: str | list[str] = None,
        group_key: str = None,
        **kwargs
    ):

        self.data = data
        self.cross_validator = cross_validator

        if isinstance(label_keys, str):
            self.labels = self.data.metadata[label_keys][:]

        else:
            # if multiple stratification labels were provided.
            self.labels = list(zip(*[self.data.metadata[i][:] for i in label_keys]))

        self.groups = self.data.metadata[group_key][:] if group_key else None

    def __iter__(self):
        self.index = 0
        splitter = self.cross_validator.split(X=self.labels, y=self.labels, groups=self.groups)

        return splitter

    def __len__(self) -> int:
        return len(self.labels)



""" Utility methods for generating diagnostic plots from loggable variables. """
from torch import Tensor

import numpy as np
from numpy.typing import ArrayLike

import pandas as pd

import matplotlib.pyplot as plt


__all__ = ("spike_raster",)


def spike_raster(
    data: Tensor, line_size: float = 0.25, events: ArrayLike | Tensor = None, event_colors: ArrayLike | Tensor = None
) -> plt.eventplot:
    """Generates a spike raster event plot from `data`.

    Parameters
    ----------
    data: Tensor
        2D binary integer tensor containing spike time series in the format units X steps.

    line_size: float
        Controls spike event line size.

    events: ArrayLike or Tensor
        Event timestamps w.r.t. X-axis. The raster will contain semi-transparent vertical bars at those locations.

    event_colors: ArrayLike or Tensor
        Colors to use with vertical lines corresponding to event timestamps, given as a list where
        the i-th element is the color of the i-th vertical line.

    """
    if data.max() <= 1:
        # we are dealing with binary spike data and need to transform it to spike times.
        # assume shape is neurons X time.
        df = pd.DataFrame(data)

        # create series of indices containing positions for raster plot.
        positions = df.apply(lambda x: df.index[x == 1.0])

        # pandas index returns inconsistent output depending on data frame content (e.g., if all rows are identical,
        # returns frame instead of list[Series]), and `plt.eventplot()` is very particular about 2D input format.
        if type(positions) is not pd.Series:
            positions = [i for i in positions.items()]
            positions = np.array(positions, dtype=object)[:, 1]

        else:
            # exit if `positions` is an empty series, as there is no need to plot an empty raster.
            if positions.empty:
                return None

        # Create raster plot with inverted y-axis to display columns in ascending order.
        plot = plt.eventplot(np.transpose(np.array(positions)[:]), linelengths=line_size, colors="black")
        # plt.yticks(range(data.shape[1]))
        plt.yticks([])

        # create vertical bars to represent events, if applicable.
        if events is not None:
            colors = [0] * len(events) if event_colors is None else event_colors
            for i, event in enumerate(events):
                plt.axvline(x=event, color=colors[i], alpha=0.5)

        plt.gca().invert_yaxis()

    else:
        plot = plt.eventplot(data, linelengths=line_size)

    # decorate plot and set axis limits.
    plt.title("Ensemble Spiking Activity")

    plt.xlabel("Time")
    plt.ylabel("Neuron")
    plt.xlim(0, data.shape[0])

    return plot



""" Constants to be used throughout the project. """
DT = 1.0
""" Global simulation time step (milliseconds). """

PLOT_BACKEND = "Agg"
""" Default matplotlib backend. """

SEED = 154
""" Random number generation (RNG) seed value to be used. Invoke :meth:`~utils.seed.set_seed` to fix the value
across all relevant libraries, e.g. numpy and scipy, before running a simulation. """

SYNAPSE_SPLITTERS = ["--", "->"]
""" Valid splitting characters for synapse files and name fields. For example, "L1->L2" and "L1--L2" are
allowed by default, to accommodate Windows' prohibition of the character ">" in filenames. """

TIME_FORMAT = "%Y%m%d-%H%M_%S"
""" Timestamp format for run directory I/O operations. """



""" Utility classes and functions for artificial signal generation. """
import logging

import torch
from torch import Tensor

from sapicore.utils.constants import DT

__all__ = ("Wave", "WaveIterable", "extend_input_current")


class WaveIterable:
    """Implicit iterator returned by the :class:`Wave` iterable.

    I.e. when you do ``iter(wave)`` or ``for x in wave:...`` it implicitly
    creates an instance of this class and uses it to iterate over.

    Parameters
    ----------
    wave: :class:`Wave:`
        The wave class that will be iterated over.
    """

    def __init__(self, wave):
        self.index = 0
        self.wave = wave

    def __next__(self):
        value = self.wave._next(self.index)
        self.index += 1
        return value


class Wave:
    """Waves are iterables that produce the next value of a specified combination of sine waves, as needed.
    This allows for the generation of arbitrary waveforms with no concern for memory usage.

    Supports basic phase-amplitude coupling as well as phase shifting frequency components with respect to each other.

    Parameters
    ----------
    amplitudes: Tensor
        Amplitudes of each frequency component, in arbitrary units.

    frequencies: Tensor
        Sine component frequencies, in Hz.

    phase_shifts: Tensor
        Starting phase of each frequency component, to be multiplied by np.pi.

    phase_amplitude_coupling: Tensor, optional
        Frequency of variations in amplitude of each component, if such coupling is desired. Defaults to `None`.

    baseline_shift: Tensor, optional
        Value by which to shift the entire signal (on the y-axis).

    sampling_rate: int, optional
        Sampling rate in Hz. Defaults to 1000.0 / :attr:`~utils.constants.DT` (simulation time step).

    device: str, optional
        Specifies a hardware device on which to process tensors, e.g. "cuda:0". Defaults to "cpu".

    See Also
    --------
    :class:`~engine.neuron.analog.oscillator.OscillatorNeuron`:
        An analog neuron that uses a Wave iterator to generate its output.

    Example
    -------
    Define a fast oscillation amplitude-coupled to the phase of a slower oscillation, maxing out at its trough:

        >>> sampling_rate = 1000
        >>> amplitudes = torch.tensor([0.1, 1.0])
        >>> frequencies = torch.tensor([50.0, 5.0])
        >>> phase_shifts = torch.tensor([torch.pi, 0.0])
        >>> phase_amplitude_coupling = torch.tensor([frequencies[1], 0.0])

    Prepare a signal based on the above parameters:

        >>> signal = Wave(amplitudes=amplitudes, frequencies=frequencies, phase_shifts = phase_shifts,
        ... phase_amplitude_coupling=phase_amplitude_coupling, sampling_rate=sampling_rate)

    Generate and plot two seconds of the signal:

        >>> import matplotlib.pyplot as plt
        >>> values = [next(signal) for _ in range(sampling_rate * 2)]
        >>> fig = plt.plot(torch.tensor(values))
        >>> plt.show()

    """

    def __init__(
        self,
        amplitudes: Tensor,
        frequencies: Tensor,
        phase_shifts: Tensor,
        phase_amplitude_coupling: Tensor = None,
        baseline_shift: Tensor = None,
        sampling_rate: float = 1000.0 / DT,
        device: str = "cpu",
    ):
        # instance attributes defining the waveform.
        self.amplitudes = amplitudes
        self.frequencies = frequencies
        self.phase_shifts = phase_shifts
        self.phase_amplitude_coupling = phase_amplitude_coupling
        self.baseline_shift = baseline_shift
        self.sampling_rate = sampling_rate

        # hardware device on which to store this instance's tensors.
        self.device = device

        # iteration counter.
        self.index = 0

        # validate correctness of waveform configuration provided.
        self._validate()

    def __iter__(self):
        # validate correctness in case params were manually changed
        self._validate()

        return WaveIterable(self)

    def _next(self, index):
        # compute wave at current time point from given components.
        value = torch.zeros(1, dtype=torch.float, device=self.device)

        for i in range(len(self.frequencies)):
            amp_series = self.amplitudes[i]
            time = index * (1.0 / self.sampling_rate)

            if self.phase_amplitude_coupling is not None:
                # add sine variation to base amplitude.
                temp = self.amplitudes[i] * torch.sin(
                    2 * torch.pi * self.phase_amplitude_coupling[i] * time + self.phase_shifts[i]
                )
                amp_series = amp_series + temp

            # add PAC oscillation to combined wave.
            value = value + amp_series * torch.sin(2.0 * torch.pi * self.frequencies[i] * time + self.phase_shifts[i])

        return value + self.baseline_shift

    def _validate(self):
        """Validates wave parameters, setting :attr:`valid` to False if configuration is faulty.
        Specifically, checks that all sine components have an amplitude, a frequency, and a phase
        (i.e., that lengths are equal)."""
        lengths = list(map(len, (self.amplitudes, self.frequencies, self.phase_shifts)))
        if not all(x == lengths[0] for x in lengths):
            raise ValueError("Not all parameters are the same length")


def extend_input_current(
    blueprint: Tensor, num_steps: int, num_units: int, device: str = "cpu", mode: str = "smear"
) -> Tensor:
    """Accepts a tensor of values `blueprint` and either smears or repeats it, depending on `mode`.

    Convenience function for generating a dummy input current from a `blueprint`. That is to say,
    the user provides the desired **shape** of the current time series and specifies how to extend it.

    For instance, if mode is "smear", [0, 1, 0, 2] will be extended to [0, 0, 1, 1, 0, 0, 2, 2, ...]
    to fit the requested duration in simulation step units, `num_steps`.  If mode is "repeat",
    [0, 1, 0, 2] will be extended to [0, 1, 0, 2, 0, 1, 0, 2, ...].

    Parameters
    ----------
    blueprint: Tensor
        Input current values to be extended (typically given under simulation/current in the configuration YAML).

    num_steps: int
        Length of current to be generated, equal to the number of simulation steps.
        This is also the number of columns in the returned tensor.

    num_units: int
        Number of neurons (tensor elements) in the ensemble that will receive the generated current.
        This is also the number of rows in the returned tensor.

    device: str, optional
        Specifies a hardware device on which to register PyTorch buffer fields, e.g. "cuda:0". Defaults to "cpu".

    mode: str, optional
        Whether to "smear" the given blueprint or "repeat" it (as in np.tile).

    Returns
    -------
    Tensor
        The generated current, given as a tensor whose columns are its values at each time point (ensemble elements
        correspond to rows).

    Examples
    --------

        >>> extend_input_current(blueprint=torch.tensor([0, 1, 1, 0]), num_steps=8, num_units=2, mode="smear")
        tensor([[0., 0., 1., 1., 1., 1., 0., 0.],
                [0., 0., 1., 1., 1., 1., 0., 0.]])

        >>> extend_input_current(blueprint=torch.tensor([0, 1, 1, 0]), num_steps=8, num_units=2, mode="repeat")
        tensor([[0., 1., 1., 0., 0., 1., 1., 0.],
                [0., 1., 1., 0., 0., 1., 1., 0.]])

    Note
    ----
    Providing a nonsensical `num_steps` (shorter than or not divisible by the length of the blueprint)
    will result in truncating or extending the output to the nearest integer multiple of the blueprint length.

    Warning
    -------
    The present implementation supports duplication of the current time series to accommodate multiple units.
    :class:`~pipeline.simulation.SimpleSimulator` uses this function to support sending varying dummy currents
    to different ensemble elements.

    Note, however, that these mechanisms are better suited for :mod:`data.synthesis`
    and will be implemented there in a principled manner. This utility function may be deprecated or moved.

    """
    # generate random current data of shape `num_steps` by `num_units`.
    chunk = blueprint.to(device)
    num_parts = num_steps // int(chunk.shape[0])
    part_size = int(chunk.shape[0])

    # sanity check the request.
    if part_size > num_steps:
        logging.warning(
            f"[extend_input_current] Your blueprint is shorter ({part_size}) than the number of steps "
            f"requested ({num_steps}). "
        )

        # use the given blueprint as is.
        num_parts = 1

    elif bool(num_steps % part_size):
        logging.warning(
            f"[extend_input_current] The number of steps you requested ({num_steps}) is not divisible "
            f"by the length of your blueprint ({part_size}). "
        )

    current_evolution = torch.zeros(part_size * num_parts, dtype=torch.float, device=device)

    for i in range(part_size):
        for j in range(num_parts):
            if mode == "smear":
                current_evolution[j + i * num_parts] = chunk[i]
            elif mode == "repeat":
                current_evolution[i + j * part_size] = chunk[i]

    current_evolution = current_evolution.repeat((num_units, 1))
    return current_evolution



""" Starts a tensorboard server with log directory set to module/script/run, given as runtime arguments.

Modules are folders containing test scripts (e.g., network, ensemble). Scripts are test scripts (e.g., test_network).
If called without -script, "test_"+module will be used by default.

Runs are individual timestamped directories within the test script directory, given as integer indices where zero
indicates the earliest run found. If called with -run -1 or without -run, the most recent run will be used.

The following calls from sapicore/tests/scripts are equivalent, reflecting the above assumptions:

* ``python -m start_tb -module network -script test_network -run -1``
* ``python -m start_tb -module network -script test_network``
* ``python -m start_tb -module network``

User may view tensorboard content by opening a browser at localhost:6006.
"""
import os
import sys
import logging

from argparse import ArgumentParser
from sapicore.utils.io import log_settings


if __name__ == "__main__":
    parser = ArgumentParser()

    parser.add_argument(
        "-module",
        action="store",
        nargs="?",
        type=str,
        dest="module",
        help="test module directory name.",
    )
    parser.add_argument(
        "-script", action="store", type=str, nargs="?", dest="script", help="test script directory name.", default=None
    )
    parser.add_argument(
        "-run",
        action="store",
        nargs="?",
        type=int,
        dest="run",
        help="run directory serial id. Most recent one loaded by default.",
        default=-1,
    )
    args = parser.parse_args()
    if not args.script:
        args.script = "test_" + args.module

    test_root = os.path.join(os.path.dirname(sys.argv[0]), "..", "engine", args.module, args.script)
    dirs = os.listdir(test_root)

    run_dirs = []
    for item in dirs:
        if item.split("-")[0].isnumeric():
            run_dirs.append(item)

    # initialize default logger.
    log_settings()

    if len(run_dirs) == 0:
        logging.info(f"No runs found at {args.module}/{args.script}. Terminating.")
    elif len(run_dirs) <= args.run:
        logging.info(f"Run ID given ({args.run}) is out of range (0-{len(run_dirs)}). Terminating.")
    else:
        # start tensorboard server at selected log directory.
        run_dirs = sorted(run_dirs)
        logging.info(f"Starting tensorboard server at {os.path.realpath(os.path.join(test_root, run_dirs[args.run]))}")

        log_dir = os.path.join(test_root, run_dirs[args.run], "tb")
        os.system(f"tensorboard --logdir={log_dir}")



""" Analog neurons emit real numbers, usually onto static synapses.

Analog neurons inherit all properties of their parent class :class:`~engine.neuron.Neuron`, extending it by
implementing a generic forward method that simply adds incoming input to their numeric state tensor `voltage`.

Analog neurons may perform normalization or provide otherwise transformed input to downstream layers.

"""
from torch import Tensor
from sapicore.engine.neuron import Neuron

__all__ = ("AnalogNeuron",)


class AnalogNeuron(Neuron):
    """Generic analog neuron endowed with a trivial :meth:`forward` method.

    Defines instance attributes shared by all derived analog neuron classes.

    Parameters
    ----------
    spiked: Tensor
        A binary representation of spike events, registered as a PyTorch buffer.

    """

    _loggable_props_: tuple[str] = ("input", "voltage")

    def __init__(self, **kwargs):
        """Invokes the parent :class:`~engine.neuron.Neuron` constructor to initialize common instance attributes."""
        super().__init__(**kwargs)

    def forward(self, data: Tensor) -> dict:
        """Updates the numeric state stored in the instance attribute tensor `voltage` to `data`.

        These default analog neurons integrate the total input impinging on them on every simulation step.


        Parameters
        ----------
        data: Tensor
            External input current to be added to this unit's numeric state tensor `voltage`.

        Raises
        ------
        RuntimeError
            If `data` tensor is not on the same hardware device as `voltage` tensor.

        Returns
        -------
        dict
            Dictionary containing the numeric state tensor `voltage`.

        """
        # update internal representation of input current for tensorboard logging purposes.
        self.input = data
        self.voltage = data

        self.simulation_step += 1

        # return current state(s) of loggable attributes as a dictionary.
        return self.loggable_state()



""" Izhikevich neuron model (IZ).

These neurons, while computationally efficient, also exhibit nontrivial dynamics resembling recorded
electrophysiological data (e.g., subthreshold oscillations).

Warning
-------
The results of the original Izhikevich simple model rely on :attr:`~utils.constants.DT` being set to 1.0 ms.
At finer temporal resolutions, weaker synchronization with a dominant frequency of ~25 Hz is observed.

See Also
--------
* `Original Izhikevich simple model <https://www.izhikevich.org/publications/spikes.pdf>`_ (Izhikevich, 2003).
* `Reproducibility <https://www.frontiersin.org/articles/10.3389/fninf.2018.00046/full>`_ (Pauli et al., 2018).

"""
import torch
from torch import tensor, Tensor

from sapicore.engine.neuron.spiking import SpikingNeuron


class IZNeuron(SpikingNeuron):
    """Base class for the Izhikevich neuron model.

    Extends :class:`~engine.neuron.spiking.SpikingNeuron` by adding necessary parameters and implementing
    the IZ forward method efficiently for 1D tensors of arbitrary size.

    User may implement subclasses to differentiate IZ model variants along additional dimensions
    (e.g., parameterizing ODE scalars as well).

    IZ units own the following attributes on top of those inherited from :class:`~engine.neuron.spiking.SpikingNeuron`:

    Parameters
    ----------
    volt_peak: float or Tensor
        Spike peak voltage, defaults to 35.0.

    u_curr: float or Tensor
        Recovery variable, defaults to -14.0.

    a: float or Tensor
        Timescale of the recovery variable `u_curr`, smaller is slower, defaults to 0.02.

    b: float or Tensor
        Sensitivity of the recovery variable `u_curr` to subthreshold fluctuations in membrane potential, defaults to
        0.2. Higher values couple `u_curr` and `voltage` more strongly, resulting in subthreshold oscillations and/or
        LTS-like dynamics.

    c: float or Tensor
        Post-spike reset value of the membrane potential, defaults to -65.0.

    d: float or Tensor
        Post-spike reset of the recovery variable `u_curr`, defaults to 2.0.

    Note
    ----
    Parameters are named `a`, `b`, `c`, and `d` in the original publication (Izhikevich, 2003).

    """

    _config_props_: tuple[str] = ("a", "b", "c", "d", "volt_peak")

    def __init__(self, a=0.02, b=0.2, c=-65.0, d=2.0, u_curr=-14.0, volt_peak=35.0, **kwargs):
        """Instantiates a single Izhikevich neuron with its default configurable and simulation attributes."""
        # register universal attributes inherited from `SpikingNeuron`, `Neuron`, and `Component`.
        super().__init__(**kwargs)

        # configurable attributes specific to Izhikevich neurons, over and above `SpikingNeuron`.
        self.volt_peak = torch.zeros(1, dtype=torch.float, device=self.device) + volt_peak
        self.a = torch.zeros(1, dtype=torch.float, device=self.device) + torch.as_tensor(a, device=self.device)
        self.b = torch.zeros(1, dtype=torch.float, device=self.device) + torch.as_tensor(b, device=self.device)
        self.c = torch.zeros(1, dtype=torch.float, device=self.device) + torch.as_tensor(c, device=self.device)
        self.d = torch.zeros(1, dtype=torch.float, device=self.device) + torch.as_tensor(d, device=self.device)

        # state attribute initialization.
        self.input = torch.zeros_like(self.c)
        self.spiked = torch.zeros_like(self.c)
        self.voltage = self.c.detach().clone()  # initialize membrane potential to resting potential.
        self.u_curr = torch.zeros_like(self.c) + u_curr

        # difference equations of this neuron model, to be used with numeric approximation methods.
        # note that voltage is the variable and u_curr a constant in eq. 1, and vice versa for eq. 2.
        self.equation = (
            lambda x, data, u: (0.04 * x + 5.0) * x + 140.0 - u + data,
            lambda x, v: self.a * (self.b * v - x),
        )

    def integrate(self, variable: str, **kwargs) -> Tensor:
        """Handles the multi-function integration necessary for IZ neurons (`u_curr` and `voltage`)."""
        if variable == "voltage":
            return self.integrator(x=self.voltage, equation=self.equation[0], **kwargs)

        else:
            return self.integrator(x=self.u_curr, equation=self.equation[1], **kwargs)

    def forward(self, data: Tensor) -> dict:
        """Izhikevich forward method, to be used by ensemble or singleton 1D tensors.

        Parameters
        ----------
        data: Tensor
            External input to be processed by this Izhikevich unit.

        Returns
        -------
        dict
            Dictionary containing the tensors `voltage`, `spiked`, and `input`.

        """
        # update internal representation of input current for tensorboard logging purposes.
        self.input = tensor([data.detach().clone()]) if not data.size() else data.detach().clone()

        # figure out which neurons spiked in the previous iteration.
        spike_fired = self.voltage >= self.volt_peak

        # compute membrane potential deltas for all neurons and update those that did not spike.
        self.voltage = spike_fired.int() * self.voltage + (~spike_fired).int() * self.integrate(
            "voltage", data=data, u=self.u_curr
        )

        # clamp membrane potential to threshold if now exceeds it.
        threshold_exceeded = self.voltage >= self.volt_peak
        self.voltage = self.voltage + threshold_exceeded * (self.volt_peak - self.voltage)

        # compute recovery variable deltas for all neurons and update those that did not spike.
        self.u_curr = spike_fired.int() * self.u_curr + (~spike_fired).int() * self.integrate("u_curr", v=self.voltage)

        # if threshold reached last iteration (spike_fired), record a spike, then reset `voltage` and `u_curr`.
        self.spiked = spike_fired

        self.voltage = self.voltage + spike_fired.int() * (self.c - self.voltage)
        self.u_curr = self.u_curr + spike_fired.int() * self.d

        self.simulation_step += 1

        # return current state(s) of loggable attributes as a dictionary.
        return self.loggable_state()



""" Oscillators are waveform generators.

Oscillators, inheriting from :class:`~engine.neuron.analog.AnalogNeuron`, generate a wave consisting of pure sines.
Components can be phase-shifted relative to each other or have their amplitude coupled to a certain frequency.

Potential uses include the imposition of modulation or forced oscillations. Like other :class:`~engine.neuron.Neuron`
instances, oscillators maintain their numeric state in the tensor `voltage`.

Warning
-------
Oscillators should only emit to non-spiking-dependent synapses (e.g, :class:`~engine.synapse.Synapse`). If paired with
synapses whose behavior requires reading out a `spiked` attribute (e.g., :class:`~engine.synapse.STDP.STDPSynapse`),
the simulation will raise an exception.

See Also
--------
utils.signals.Wave:
    Iterator class used to generate complex waveforms without memory overhead.

"""
from math import pi
from torch import as_tensor, tensor, Tensor

from sapicore.engine.neuron.analog import AnalogNeuron
from sapicore.utils.signals import Wave

__all__ = ("OscillatorNeuron",)


class OscillatorNeuron(AnalogNeuron):
    """Oscillator neuron implementation.

    Waveform is determined by amplitude, frequency, phase, and coupling configuration parameters in a
    format acceptable to the utility class :class:`~utils.signals.Wave` constructor.

    Parameters
    ----------
    amplitudes: Tensor or list of float
        Component amplitudes in arbitrary units.

    frequencies: Tensor or list of float
        Component frequencies in Hz.

    phases: Tensor or list of float
        Component starting phases between 0 and 2, to be multiplied by pi.

    amp_freq: Tensor or list of float
        Frequencies for amplitude modulation of Nth item, e.g. if frequencies[0] = 40.0
        and amp_freq[0] = 5.0, the amplitude of the 40 Hz component will oscillate at 5 Hz.

    baseline_shift: Tensor or list of float, optional
        Value by which to shift the entire signal (on the y-axis).

    **kwargs:
        Accepts and applies any keyword argument by invoking the parent class :class:`~engine.neuron.Neuron`
        constructor.

    """

    # loggable properties are the same as those of the parent class AnalogNeuron.
    _config_props_: tuple[str] = ("amplitudes", "frequencies", "phases", "amp_freq", "baseline_shift")

    def __init__(
        self,
        amplitudes: Tensor | list[float] = None,
        frequencies: Tensor | list[float] = None,
        phases: Tensor | list[float] = None,
        amp_freq: Tensor | list[float] = None,
        baseline_shift: Tensor | list[float] = None,
        **kwargs
    ):
        """Constructs an iterator-based oscillator."""
        super().__init__(**kwargs)

        self.amplitudes = as_tensor(amplitudes if amplitudes is not None else [5.0], device=self.device)
        self.frequencies = as_tensor(frequencies if frequencies is not None else [40.0], device=self.device)
        self.phases = as_tensor(phases if phases is not None else [0.0], device=self.device)
        self.amp_freq = as_tensor(amp_freq if amp_freq is not None else [0.0], device=self.device)
        self.baseline_shift = as_tensor(baseline_shift if baseline_shift is not None else [0.0], device=self.device)

        # placeholder for waves and wave iterator.
        self._iter = []
        self._waves = []

    def register_waveforms(self) -> None:
        """Creates Wave iterators for each oscillator element."""

        # create an iterator for each oscillator in the ensemble and append it to `iter`.
        self._iter = []
        self._waves = []
        for i in range(self.amplitudes.shape[0]):
            specification = {
                "amplitudes": self.amplitudes[i],
                "frequencies": self.frequencies[i],
                # convert phase specification from multiples-of-pi to radians.
                "phase_shifts": self.phases[i] * pi,
                "phase_amplitude_coupling": self.amp_freq[i],
            }

            wave = Wave(
                **specification, sampling_rate=self.dt * 1000.0, device=self.device, baseline_shift=self.baseline_shift
            )
            self._waves.append(wave)
            self._iter.append(iter(wave))

    def forward(self, data: Tensor) -> dict:
        """Oscillator forward method.

        Advances the wave by one time unit and adds the `data` tensor.

        Parameters
        ----------
        data: Tensor
            External input current to be added to the waveform.

        Returns
        -------
        dict
            Dictionary containing the numeric state tensor `voltage`.

        """
        # compute the next value of the wave and store it in `voltage`, the state variable of the analog neuron.
        # if input current is provided (`data` tensor), it will be added to the oscillation.
        self.voltage = tensor([next(wave_sample) for wave_sample in self._iter], device=self.device).add(data)
        self.simulation_step += 1

        # return current state(s) of loggable attributes as a dictionary.
        return self.loggable_state()



import pytest
import os

import torch
from sklearn.model_selection import StratifiedKFold, StratifiedGroupKFold

from sapicore.data import Data, AxisDescriptor
from sapicore.data.sampling import BalancedSampler, CV

from sapicore.utils.sweep import Sweep
from sapicore.tests import ROOT

from sapicore.engine.network import Network
from sapicore.model import Model

TEST_ROOT = os.path.join(ROOT, "tests", "data", "test_data")  # test root directory.


class TestData:
    @pytest.mark.unit
    def test_data_pipeline(self):
        # initialize a dummy dataset with balanced labels.
        data = Data(buffer=torch.rand(8, 4))
        params = Sweep({"grid": {"study": [1, 2], "animal": ["A", "B", "C", "D"]}}, 8)(dataframe=True)

        study = AxisDescriptor(name="study", labels=params["study"].tolist(), axis=0)
        animal = AxisDescriptor(name="animal", labels=params["animal"].tolist(), axis=0)
        sensor = AxisDescriptor(name="sensor", labels=["Occipital", "Parietal", "Temporal", "Frontal"], axis=1)

        # register the above axis descriptors (label vectors) with our Data object.
        data.metadata.add_descriptors(study, animal, sensor)

        # print the row descriptors in tabular form.
        data.metadata.to_dataframe(axis=0)

        # different ways of accessing label vectors and data with slicing.
        _ = data.metadata["sensor"][:]
        _ = data[:]

        # demonstrate stratified sampling w.r.t. one descriptor using a base cross validator.
        data.sample(method=StratifiedKFold, shuffle=True, retain=4, label_keys="animal")

        # demonstrate balanced sampling w.r.t. two descriptors using a BalancedSampler (1 for each crossing).
        data.sample(method=BalancedSampler(replace=False, stratified=False), group_keys=["animal", "study"], n=1)

        # set up a cross validation object, leveraging scikit-learn.
        cv = CV(
            data=data, cross_validator=StratifiedGroupKFold(2, shuffle=True), label_keys="animal", group_key="study"
        )

        # container for models trained in each CV fold.
        models = []

        for i, (train, test) in enumerate(cv):
            # initialize a new model for this cross validation fold and append it to the list.
            # a dummy network is used, but a real one could be initialized by providing a `configuration` dict.
            models.append(Model(Network()))

            # repeats each sample for a random number of steps (simulating variable exposure durations).
            models[i].fit(data[train], duration=torch.randint(low=2, high=7, size=(data[train].shape[0],)))

    @pytest.mark.parametrize(
        "url_",
        [
            "https://archive.ics.uci.edu/ml/machine-learning-databases/00224/Dataset.zip",
        ],
        ids=["DRIFT"],
    )
    @pytest.mark.unit
    def test_fetch(self, url_):
        Data(remote_urls=url_, root=os.path.join(TEST_ROOT, os.path.basename(url_).split(".")[0]))


if __name__ == "__main__":
    pytest.main(args=["-s", "-v"])



""" Ensemble unit tests. """
import os
import pytest

import torch

from sapicore.engine.ensemble.analog import AnalogEnsemble, OscillatorEnsemble
from sapicore.engine.ensemble.spiking import LIFEnsemble, IZEnsemble

from sapicore.tests import ROOT, TEST_DEVICE


STEPS = 100  # default number of simulation forward steps to execute during testing.
TEST_ROOT = os.path.join(ROOT, "tests", "ensemble", "test_ensemble")  # test root directory.


class TestEnsemble:
    @pytest.mark.parametrize(
        "arg_",
        [
            AnalogEnsemble(device=TEST_DEVICE),
            AnalogEnsemble(num_units=1000, device=TEST_DEVICE),
        ],
        ids=["DEFAULT", "MULTIUNIT"],
    )
    @pytest.mark.unit
    def test_analog(self, arg_: AnalogEnsemble | OscillatorEnsemble):
        """Task initialization test cases."""
        # verify that all attributes are initialized.
        assert all(hasattr(arg_, attr) for attr in ["identifier", "input", "voltage", "num_units"])

        # verify that loggable attributes are torch tensors on the correct device.
        assert all(type(getattr(arg_, attr)) is torch.Tensor for attr in arg_._loggable_props_)
        assert all(getattr(arg_, attr).device == torch.device(TEST_DEVICE) for attr in arg_._loggable_props_)

        # verify behavior of trivial forward method.
        init_voltage = arg_.voltage

        # adding a compatible shape data should increase `voltage` buffer values.
        add_same_shape = arg_.forward(data=torch.ones_like(init_voltage, device=TEST_DEVICE))
        assert torch.all(add_same_shape["voltage"].eq(init_voltage + 1))
        assert torch.all(add_same_shape["voltage"].eq(arg_.voltage))

    @pytest.mark.parametrize(
        "arg_",
        [
            LIFEnsemble(device=TEST_DEVICE),
            IZEnsemble(num_units=1000, device=TEST_DEVICE),
        ],
        ids=["LIF", "IZ"],
    )
    @pytest.mark.unit
    def test_generic(self, arg_: LIFEnsemble | IZEnsemble):
        # TODO complete unit tests and add cases as necessary.
        pass


if __name__ == "__main__":
    pytest.main(args=["-s", "-v"])



""" Run this script to purge old run output and caches recursively in all test script directories. """
import os
import shutil
import logging

from sapicore.utils.io import log_settings


def get_size(start_path: str = "."):
    """Recursively find total size in bytes of files in or below the directory `start_path`."""
    total_size = 0
    for dir_path, dir_names, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dir_path, f)

            # skip if it is symbolic link.
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size


if __name__ == "__main__":
    # initialize default logger.
    log_settings()

    # initialize list of directories to remove.
    chopping_block = []

    # if script directory treated as root, go up to project root.
    src_root = os.path.join(os.getcwd(), "..", "..") if "scripts" in os.getcwd() else os.getcwd()

    # walk all directories below project root recursively and delete test run output, pycache, and coverage.
    for root, _, _ in os.walk(src_root):
        items = os.listdir(root)
        for item in items:
            if os.path.basename(root).startswith("test_"):
                if item.split("-")[0].isnumeric():
                    logging.info(f"Deleting run {item} files from {os.path.basename(root)}.")
                    chopping_block.append(os.path.join(root, item))
            else:
                deletion = any([word in item for word in ["coverage", "pytest", "pycache", "hypothesis"]])
                exclusion = ".ini" in item or "coveragerc" in item

                if deletion and not exclusion:
                    # also removes code coverage report and caches in all test subdirectories.
                    logging.info(f"Deleting {item} from {os.path.basename(root)}.")
                    chopping_block.append(os.path.join(root, item))

    # space to be freed in MB.
    space = sum([get_size(item) for item in chopping_block]) / 10**6

    for item in chopping_block:
        try:
            shutil.rmtree(item) if os.path.isdir(item) else os.remove(item)
        except FileNotFoundError:
            pass

    logging.info(f"Freed {space:.2f} MB.")



"""UCSD Drift dataset."""
from zipfile import ZipFile

import os
import shutil

import torch
import pandas as pd

from scipy.sparse import csr_matrix as csr
from sklearn.datasets import load_svmlight_file

from sapicore.data import Data, AxisDescriptor, Metadata
from sapicore.utils.io import ensure_dir


class DriftDataset(Data):
    """Processing logic for the UCSD drift dataset.

    Implements the protected :meth:`~data.Data._standardize` to extract, transform, and save buffer and labels
    from the raw .dat files downloaded from the UCI repository.

    See Also
    --------
    `Drift dataset specification <https://archive.ics.uci.edu/dataset/224/gas+sensor+array+drift+dataset>`_

    """

    def __init__(self, root: str, **kwargs):
        super().__init__(
            root=root,
            download=True,
            remote_urls="https://archive.ics.uci.edu/static/public/224/gas+sensor+array+drift+dataset.zip",
            **kwargs
        )

    def _standardize(self):
        # the file retrieved from the UCSD repository needs to be unzipped.
        with ZipFile(os.path.join(self.root, "gas+sensor+array+drift+dataset.zip"), "r") as zf:
            zf.extractall(path=os.path.join(self.root))

        # remove zip archive and move extracted files to the dataset root directory "drift".
        os.remove(os.path.join(self.root, "gas+sensor+array+drift+dataset.zip"))
        for file in os.listdir(os.path.join(self.root, "Dataset")):
            shutil.move(os.path.join(self.root, "Dataset", file), os.path.join(self.root, file))
        shutil.rmtree(os.path.join(self.root, "Dataset"))

        # scan the root directory for .dat files, which happen to be in libsvm format.
        raw_files = self.scan(pattern="*.dat")

        # placeholder for accumulating label vectors.
        meta = {"chemical": [], "batch": []}

        for i, file in enumerate(raw_files):
            # load and convert from sparse matrix to standard numpy array.
            data, labels = load_svmlight_file(file)
            data = csr(data, dtype=data.dtype).todense()

            # since the dataset is small, vertically stack batch-specific data onto the `buffer`.
            self.buffer = (
                torch.tensor(data, dtype=torch.float64)
                if i == 0
                else torch.vstack((self.buffer, torch.tensor(data, dtype=torch.float64)))
            )

            meta["chemical"].extend(labels)
            meta["batch"].extend([i + 1 for _ in range(len(labels))])

        # add metadata to dataset.
        self.metadata = Metadata(
            AxisDescriptor(name="chemical", labels=meta["chemical"], axis=0),
            AxisDescriptor(name="batch", labels=meta["batch"], axis=0),
        )

        # dump the data and label tensors to .pt for loading convenience.
        self.save()

    def load(self, indices: slice = None):
        """Load drift data and labels to memory from the .pt files saved by :meth:`~data.Data._standardize`.
        Overwrites `buffer` and `descriptors`. If specified, selects only the rows at `indices`.

        Parameters
        ----------
        indices: slice, optional
            Specific indices to load.

        """
        # the labels file has been created by _standardize upon first obtaining the data.
        labels = torch.load(os.path.join(self.root, "labels.pt"), weights_only=False)
        if indices is not None:
            labels = labels[indices]

        # for the drift set specifically, we know that all labels describe the 0th axis (rows).
        self.metadata = Metadata(*[AxisDescriptor(name=col, labels=labels[col].to_list(), axis=0) for col in labels])

        # load the data into the buffer (since the drift set is small, no need for lazy loading).
        self.buffer = torch.load(os.path.join(self.root, "data.pt"), weights_only=False)

        if indices is not None:
            self.buffer = self.buffer[indices]

        return self

    def save(self):
        """Dump the contents of the `buffer` and `metadata` table to disk at `destination`.
        This implementation saves the tensors as .pt files.

        """
        torch.save(self.buffer, os.path.join(ensure_dir(self.root), "data.pt"))
        torch.save(self.metadata.to_dataframe(), os.path.join(ensure_dir(self.root), "labels.pt"))

    def select(self, conditions: str | list[str], axis: int = 0) -> slice:
        """Selects sample indices by descriptor values based on a pandas logical statement applied to one or more
        :class:`AxisDescriptor` objects, aggregated into a dataframe using :meth:`aggregate_descriptors`.

        Importantly, these filtering operations can be performed prior to loading any data to memory, as they only
        depend on :class:`AxisDescriptor` objects (labels) attached to this dataset.

        Note
        ----
        Applying selection criteria with pd.eval syntax is simple:

        * "age > 21" would execute `pd.eval("self.table.age > 21", target=df)`.
        * "lobe.isin['F', 'P']" would execute `pd.eval("self.table.lobe.isin(['F', 'P']", target=df)`.

        Where `df` is the table attribute of the :class:`~AxisDescriptor` the condition is applied to.

        Warning
        -------
        This method is general and may eventually be moved to the base :class:`~data.Data` class.

        """
        # wrap a single string condition in a list if necessary.
        if isinstance(conditions, str):
            conditions = [conditions]

        # aggregate relevant axis descriptors into a pandas dataframe (table).
        df = self.metadata.to_dataframe(axis=axis)

        # evaluate the expressions and filter the dataframe.
        for expression in conditions:
            parsed = "& ".join([("df." + expr).replace(". ", ".") for expr in expression.split("&")])
            parsed = "| ".join([("df." + expr).replace(". ", ".") for expr in parsed.split("|")])

            parsed = parsed.replace("df.df.", "df.")
            parsed = parsed.replace("df.(", "(df.")

            df = df[pd.eval(parsed, target=df).to_list()]

        # return indices where expressions evaluated to true.
        return df.index.to_list()



""" Sweep hyperparameter or argument search spaces. """
from typing import Any
from torch import Tensor

import torch
import scipy

import numpy as np
import pandas as pd

from sklearn.model_selection import ParameterGrid


class Sweep:
    """Reads a `search_space` dictionary and generates appropriate parameter combinations.

    Parameters
    ----------
    search_space: dict
        Dictionaries of argument keys and their possible values (or scipy distributions and parameters),
        grouped by top-level keys specifying the sweep strategy to be used with each.
        Search spaces are typically stored in the `sweep` key  of a :class:`~engine.component.Component`
        configuration dictionary.


    Search space parameters are expected to be found under one of four top-level keys:

    1. **Fixed**: The argument takes the same value across all branches. E.g., if
    settings["fixed"]["resample"]=100, resample will always take the value 100.

    2. **Zipped**: The values of the arguments listed are iterated over such that the Nth value of any argument is
    matched with the Nth value of other arguments. Combinations are repeated as many times as it takes
    to reach the number of combinations dictated by construction or by specification.
    E.g., if settings["zipped"]["freq_band"] = [[5, 10], [30, 50]] and ... ["resample"] = [50, 100],
    then `self.combinations` will have freq_band = [5, 10] with resample = 50 in half the branches,
    and freq_band = [30, 50] with resample = 100 in the other half.

    3. **Grid**: The values of the keys listed are iterated over such that any value combination is
    equally represented in the resulting dictionary. Combinations are repeated as many times as it takes to reach
    the number of combinations dictated by construction or by specification.
    E.g., if settings["zipped"]["freq_band"] = [[5, 10], [30, 50]] and ...["resample"] = [50, 100],
    then `self.combinations` will have four qualitatively distinct groups, with (freq_band, resample) taking
    one of the following combined values: ([5, 10], 50), ([5, 10], 100), ([30, 50], 50), ([30, 50], 100).

    4. **Random**: The values of the parameter keys listed are drawn from a :mod:`scipy.stats` distribution given
    under the key "method", with the arguments "args". E.g., if settings["random"]["resample"]["method"]="uniform"
    and ...["args"] = [x, y], then `resample` values will be drawn from `U(x, x+y)`.

    Example
    -------
    Define a search space and generate a list of parameter combination dictionaries:

        >>> s = Sweep(search_space={"fixed": {"a": 4}, "grid": {"b": [6, 7], "c": [8, 9]}}, num_combinations=4)
        >>> print(s())
        [{'a': 4, 'b': 6, 'c': 8}, {'a': 4, 'b': 6, 'c': 9}, {'a': 4, 'b': 7, 'c': 8}, {'a': 4, 'b': 7, 'c': 9}]

    See Also
    --------
    `Scipy Distributions <https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions>`_
        For a list of supported probability distributions and their arguments.

    """

    def __init__(self, search_space: dict, num_combinations: int = None):
        self.search_space = search_space
        self.num_combinations = num_combinations

        # generate argument combinations from `search_space`.
        self.combinations = self.generate_combinations()

    def __call__(self, dataframe: bool = False):
        """Calling the object will return the combinations as either a list of dictionaries (default)
        or a pandas dataframe."""
        if dataframe:
            return pd.DataFrame(self.combinations)

        else:
            return self.combinations

    def heterogenize(self, obj: Any, unravel: bool = True):
        """Initializes attribute tensors heterogeneously based on a given sweep search dictionary.

        Parameters
        ----------
        obj: Component
            Sapicore engine component whose parameters will be diversified (currently, either
            :class:`~engine.ensemble.Ensemble` or :class:`~engine.synapse.Synapse`).

        unravel: bool
            When the property is a 2D tensor, dictates whether combination values should be assigned
            to individual elements (True, default) or overwrite entire rows (False).

        """
        # for every parameter combination, set the appropriate attribute tensor element values.
        for i, combination in enumerate(self.combinations):
            for key, value in combination.items():
                # cast attribute value to tensor on the correct device.
                value = torch.tensor(value, device=obj.device)

                # some attributes may not exhaust all possible combinations (e.g., if some are 2D and some are 1D).
                attr = getattr(obj, key)
                if isinstance(attr, Tensor) and i < attr.numel():
                    # FIX assumes attribute to be modified is already a tensor after super().__init__.
                    if unravel:
                        # compute unraveled index given the shape of the tensor to be heterogenized.
                        unraveled_index = np.unravel_index(i, attr.shape)
                        attr[unraveled_index] = value

                    else:
                        # handle the case where values should overwrite attribute tensor rows.
                        attr[i] = value
                else:
                    attr = value

    def generate_combinations(self) -> list[dict]:
        """Determine the argument combinations defining this sweep, update the `combinations` attribute,
        and return its value for potential use by the calling function."""
        # allow user to set an arbitrary number of combinations, e.g. in the case of random sampling.
        # if not specified, `count_combinations` will use the LCM of grid and zipped combinations to set the total.
        self.count_combinations()

        # initialize an empty list to contain kwargs-like dictionaries, one for each branch/combination.
        combinations = []

        # combinations to cycle through when setting zipped argument values.
        zipped_cycle = self.extract_combinations(self.search_space, mode="zipped")

        # combinations to cycle through when setting grid argument values.
        grid_cycle = self.extract_combinations(self.search_space, mode="grid")

        for n in range(self.num_combinations):
            # initialize temporary array buffer to store combination dictionary from this iteration.
            combinations.append({})

            # handle the fixed parameters first.
            if "fixed" in self.search_space:
                for arg in self.search_space["fixed"].keys():
                    combinations[n][arg] = self.search_space["fixed"][arg]

            # handle the zipped parameters next, interleaving configurations.
            if "zipped" in self.search_space:
                for arg in self.search_space["zipped"].keys():
                    combinations[n][arg] = zipped_cycle[int(n % len(zipped_cycle))][arg]

            # handle the grid-search parameters next, interleaving configurations.
            if "grid" in self.search_space:
                for arg in self.search_space["grid"].keys():
                    combinations[n][arg] = grid_cycle[int(n % len(grid_cycle))][arg]

        # finally, draw random parameter values from their scipy distributions.
        if "random" in self.search_space:
            for arg in self.search_space["random"].keys():
                random_distribution = getattr(scipy.stats, self.search_space["random"][arg]["method"])
                random_variable = random_distribution(*self.search_space["random"][arg]["args"])
                random_values = random_variable.rvs(size=self.num_combinations)

                for n in range(self.num_combinations):
                    combinations[n][arg] = random_values[n]

        return combinations

    def count_combinations(self):
        """Counts search space configurations and sets `num_combinations` if not provided or invalid.

        Enforces the following contract w.r.t. the calculated number of combinations (NoC):

        1. If NoC was not provided by the user, the calculated value is the minimal one ensuring that every unique
        parameter combination can be represented.

        2. If NoC was provided, the calculated value will be either equal to or greater than the requested number.
        The calculated value will be greater iff the user's value will result in an imbalance (i.e., if not
        every parameter combination can be represented with that NoC). In such cases, the user should
        intentionally sample from the larger combination dictionary in a way they see fit. Generally,
        users should keep balancing considerations in mind when designing heterogeneous layers.

        """
        zipped_combinations = 0
        grid_combinations = 0

        for key, value in self.search_space.items():
            if key == "zipped" and self.search_space["zipped"]:
                # if a nonempty "zipped" strategy exists in the configuration, we have as many additional
                # branches to add as the common length of the zipped parameter options.
                zipped_combinations = len(list(self.search_space["zipped"].values())[0])

            if key == "grid" and self.search_space["grid"]:
                # use the scikit-learn utility class ParameterGrid to extract possible combinations.
                grid_combinations = len(list(ParameterGrid(self.search_space["grid"])))

        # determine number of unique combinations if one was not requested.
        lcm = np.lcm(zipped_combinations, grid_combinations)
        if not self.num_combinations:
            if zipped_combinations + grid_combinations == 0:
                # if we got here without zipped and grid combos, this means user only had fixed variables,
                # or had random variables but didn't set `num_combinations`. In that case, default to 1.
                self.num_combinations = 1

            elif lcm == 0:
                # if only zipped or only grid variables, whichever is nonzero is the number of combinations.
                self.num_combinations = zipped_combinations if zipped_combinations else grid_combinations

        elif self.num_combinations < lcm:
            # if `num_combinations` provided is smaller than the number of unique combinations,
            # set it to the least common multiplier to ensure balanced output.
            self.num_combinations = lcm

    @staticmethod
    def extract_combinations(parameters: dict, mode: str) -> list[dict]:
        """Extracts all parameter combinations from `parameters` based on the traversal `mode` given.

        Parameters
        ----------
        parameters: dict
            Dictionary, possibly loaded from YAML, containing parameters and their levels.

        mode: str
            Combination strategy, "grid" if all parameter permutations should be tested or "zipped" for a linear pass.

        See Also
        --------
        :class:`~utils.sweep.Sweep`:
            For a full documentation of parameter sweep strategies.

        """
        combinations = []

        # return list of dictionaries with all possible parameter permutations.
        if mode == "grid":
            if "grid" in parameters:
                parameters = parameters["grid"]
                param_grid = ParameterGrid(parameters)

                for dict_ in param_grid:
                    combinations.append(dict_)

        # return list of dictionaries with serially generated parameter permutations (key_1[n] goes with key_2[n]).
        elif mode == "zipped":
            if "zipped" in parameters:
                parameters = parameters["zipped"]

                keys = list(parameters)

                # verify that, in zipped mode, all parameters have the same number of levels.
                if len(keys) > 1:
                    if not [len(parameters[keys[i]]) == len(parameters[keys[i + 1]]) for i in range(len(keys) - 1)]:
                        raise AssertionError("Different number of levels for one or more parameters.")

                for i in range(len(parameters[keys[0]])):
                    # make copy of full dictionary and remove all levels except i.
                    combinations.append(parameters.copy())
                    for j in keys:
                        combinations[i][j] = combinations[i][j][i]

        else:
            raise ValueError("Invalid parameter combination mode provided.")

        return combinations



""" Data operations. """
from typing import Any, Callable, Optional, Sequence
from numpy.typing import NDArray

import os
import urllib.request
from urllib.error import HTTPError, URLError

from copy import deepcopy
from glob import glob
from natsort import natsorted

import torch
import numpy as np
import pandas as pd

from torch import Tensor
from torch.utils.data import Dataset

from sklearn.model_selection import BaseCrossValidator
from sklearn.preprocessing import normalize as norm

from sapicore.data.sampling import BalancedSampler, CV
from sapicore.utils.io import ensure_dir


__all__ = ("AxisDescriptor", "Metadata", "Data")


class AxisDescriptor:
    """Metadata vector of `labels` describing a particular axis of an N-dimensional array.

    Parameters
    ----------
    name: str, optional
        Name of the descriptor (i.e., the variable whose value is given by the labels).

    labels: list or NDarray or Tensor, optional
        Label values given as a list, numpy array, or tensor. Regardless of passed type, these are internally
        converted to a numpy array (not tensors, as those do not support string labels).

    axis: int, optional
        The data axis described by this variable. Defaults to 1 (columns).

    Example
    -------
    Generate a dummy dataset and describe its axes with multiple lists of labels:

        >>> import torch
        >>> data = Data(buffer=torch.rand(8, 4))
        >>> study = AxisDescriptor(name="study", labels=[1, 1, 2, 2, 1, 1, 2, 2], axis=0)
        >>> animal = AxisDescriptor(name="animal", labels=["A", "A", "A", "A", "B", "B", "B", "B"], axis=0)
        >>> sensor = AxisDescriptor(name="sensor", labels=["Occipital", "Parietal", "Temporal", "Frontal"], axis=1)

    Here, we have eight buffer (rows) of four measured dimensions (columns). The third AxisDescriptor indicates
    that the columns correspond to sensor locations. The first two describe the rows, and contain information
    about the study and animal each sample was obtained from.

    """

    def __init__(self, name: str = "", labels: list | NDArray | Tensor = None, axis: int = 0):
        self.name = name
        self.axis = axis
        self.labels = np.array(labels)

    def __getitem__(self, index: Any) -> Tensor:
        return self.labels[index]

    def __setitem__(self, index: Any, values: Any):
        self.labels[index] = np.array(values)


class Metadata:
    """Collection of :class:`AxisDescriptor` objects indexed in a table."""

    def __init__(self, *args: AxisDescriptor):
        # verify that no duplicate descriptors exist in args.
        self._validate_names(*args)

        # dictionary containing descriptor names as keys and their references as values.
        self.table = {descriptor.name: descriptor for descriptor in args}

    def __getitem__(self, key: str) -> AxisDescriptor:
        """Return the label vector `key` if exists."""
        if key in self.table.keys():
            return self.table[key]

        else:
            raise KeyError(f"Metadata does not contain the key {key}.")

    def __setitem__(self, key: str, value: list | NDArray | Tensor | AxisDescriptor):
        """Creates an :class:`AxisDescriptor` from the `value` vector and adds it to this instance as
        a row descriptor (axis=0).

        Note
        ----
        Guarantees that the `name` attribute of the AxisDescriptor matches its key in `table`.

        """
        if isinstance(value, AxisDescriptor):
            if value.name != key:
                raise ValueError(f"Key does not match ({key} != {value.name})")

            self.table[key] = value

        else:
            self.table[key] = AxisDescriptor(name=key, labels=np.array(value), axis=0)

    def add_descriptors(self, *args: AxisDescriptor):
        """Adds one or more :class:`AxisDescriptor` objects to this metadata instance."""
        self._validate_names(*args)

        for descriptor in args:
            self.table[descriptor.name] = descriptor

    def to_dataframe(self, axis: int = 0) -> pd.DataFrame:
        """Condenses the :class:`~data.Data.AxisDescriptor` objects registered in this dataset's
        :class:`data.Data.Metadata` `table` attribute to a pandas dataframe.

        Performed for row (sample) descriptors by default, but can also be used to aggregate
        :class:`AxisDescriptor` objects describing any other `axis`.

        Note
        ----
        If descriptors for a particular axis vary in length, NaN values will be filled in as necessary.

        """
        df = pd.DataFrame()
        for key, values in self.table.items():
            if values.axis == axis:
                df[key] = pd.Series(values.labels)

        return df

    @staticmethod
    def _validate_names(*args: AxisDescriptor):
        if len(args) != len(set(descriptor.name for descriptor in args)):
            raise ValueError(f"Got multiple descriptors with the same name in {args}")


class Data(Dataset):
    """Dataset base class.

    Provides an interface and basic default implementations for fetching, organizing,
    and representing external datasets. Designed to be incrementally extended.

    Parameters
    ----------
    identifier: str, optional
        Name for this dataset.

    metadata: Metadata, optional
        Metadata object describing this dataset. Consists of multiple :class:`AxisDescriptor` references.

    root: str, optional
        Local root directory for dataset file(s), if applicable.

    remote_urls: str or list of str, optional
        Remote URL(s) from which to fetch this dataset, if applicable.

    buffer: Tensor, optional
        A buffer holding data buffer in tensor form. Useful for initializing objects on the fly, e.g.
        during data synthesis. Users may leverage :meth:`access` to manage disk I/O, e.g.
        using :class:`torch.storage.Storage`, memory mapped arrays/tensors, or HDF lazy loading.

    download: bool, optional
        Whether to download the set. Defaults to `False`.
        If `True`, the download only commences if the `root` doesn't exist or is empty.

    overwrite: bool, optional
        Whether to download the set regardless of whether `root` already contains cached/pre-downloaded data.

    labels: Sequence, optional
        Labels to create row metadata from, using a default procedure.

    See Also
    --------
    `PyTorch Storage Class <https://discuss.pytorch.org/t/memory-mapped-tensor/8954>`_

    """

    def __init__(
        self,
        identifier: str = "",
        buffer: Tensor = None,
        metadata: Metadata | Sequence = None,
        root: Optional[str] = None,
        remote_urls: str | list[str] = "",
        download: bool = False,
        overwrite: bool = False,
        labels: Sequence = None,
        **kwargs,
    ):
        # dataset identifier, sometimes used with default methods.
        self.identifier = identifier

        # optional data tensor provided by user at instantiation.
        self.buffer = buffer

        # metadata object containing axis-specific descriptor labels.
        if isinstance(metadata, Metadata):
            self.metadata = metadata
        else:
            self.metadata = Metadata()
            if isinstance(metadata, Sequence) and labels is not None:
                # if labels supplied, use them to create row metadata.
                self.metadata.add_descriptors(AxisDescriptor(labels))

        # optional, local and remote directories.
        self.root = root
        self.remote_urls = remote_urls

        # fetch dataset from remote URL if `root` doesn't exist or is empty or if asked to do so regardless.
        if overwrite or (download and not self._check_downloaded()):
            self._download()

            # converts the foreign dataset to a single file in a user-determined format.
            # passes silently if not implemented by the user.
            self._standardize()

    def __getitem__(self, index: Any):
        """Calls :meth:`access` to slice into the data or access specific file(s), returning the value(s) at `index`."""
        return self.access(index)

    def __setitem__(self, index: Any, values: Tensor):
        """Sets buffer values at the given indices to `values`."""
        self.modify(index, values)

    def __len__(self):
        """The default implementation addresses trivial cases where the set is loaded to memory (in `self.buffer`).

        Warning
        -------
        Users should override this method when dealing with large datasets that are dynamically loaded using
        a :meth:`access` call.

        """
        return len(self.buffer)

    def _check_downloaded(self) -> bool:
        """Checks whether files have already been downloaded.

        By default, returns `True` if root directory exists and is nonempty.

        Note
        ----
        Users may wish to override this behavior in their own derivative dataset-specific classes,
        e.g. to include checksum verification.

        """
        return os.path.exists(self.root) and len(os.listdir(self.root)) > 0

    def _download(self):
        """Downloads one or more files to the `root` directory from `remote_urls`.

        Raises
        ------
        TypeError
            If a URL is not provided and the data does not exist locally at `root`.

        ValueError
            If the URL provided is invalid.

        ConnectionError
            If remote URL is down.

        """
        # create the destination directory if it doesn't exist.
        ensure_dir(self.root)

        # if given a single URL, wrap it in a list.
        if isinstance(self.remote_urls, str):
            self.remote_urls = [self.remote_urls]

        # download file(s).
        for url in self.remote_urls:
            try:
                file = url.split("/")[-1]
                urllib.request.urlretrieve(url, os.path.join(self.root, file))
            except HTTPError as e:
                raise ConnectionError(f"Download failed, server down: {url}") from e
            except URLError as e:
                raise ConnectionError(f"Download failed, URL invalid: {url}") from e

    def _standardize(self):
        """Standardizes an external data directory tree, e.g. a remote repository, possibly converting it to a
        single condensed file at `root` for subsequent operations. Should only be performed once after fetching
        the data from its remote source, if applicable.

        This method is meant to be implemented by child classes that deal with arbitrarily formatted external sets.
        It may also serve as a bridge between Sapicore and specialized libraries, e.g. MNE for EEG/MEG data.

        Warning
        -------
        The resulting file may follow any convention the user would like to adopt. Sapicore makes no assumptions;
        it is the user's responsibility to implement this method and :meth:`access` below in a compatible way.

        """
        pass

    def get_metadata(self, keys: bool = True) -> list[Any]:
        """Returns metadata keys by default, or their values (AxisDescriptor references) if `keys` is False."""
        return list(self.metadata.table.keys()) if keys else list(self.metadata.table.values())

    def access(self, index: Any, axis: int = None) -> Tensor:
        """Specifies how to access data by mapping indices to actual samples (e.g., from file(s) in `root`).

        The default implementation slices into `self.buffer` to accommodate the trivial cases where the user has
        directly initialized this :class:`Data` object with a `buffer` tensor or loaded its values by reading
        a file that fits in memory (the latter case would be handled by :meth:`load`).

        More sophisticated use cases may require lazy loading or navigating HDF files. That kind of logic should
        be implemented here by derivative classes.

        Parameters
        ----------
        index: Any
            Index(es) to slice into.

        axis: int, optional
            Optionally, a specific axis along which to apply index selection.

        Note
        ----
        Where file data are concerned (e.g., audio/image, each being a labeled "sample"), use this method to read
        and potentially transform them, returning the finished product.

        """
        return self.buffer.index_select(axis, torch.as_tensor(index)) if axis is not None else self.buffer[index]

    def load(self, indices: Any = None):
        """Populates the `buffer` tensor buffer and/or `descriptors` attribute table by loading one or more files
        into memory, potentially selecting only `indices`.

        Since different datasets and pipelines call for different formats, implementation is left to the user.

        Parameters
        ----------
        indices: Any
            Specific indices to include, one for each file.

        Returns
        -------
        Data
            Self reference. For use in single-line initialization and loading.

        Warning
        -------
        Populating the `buffer` with the entire dataset should only be done when it can fit in memory.
        For large sets, the buffer should not be used naively; :meth:`access` should be overriden
        to implement some form of lazy loading.

        """
        pass

    def modify(self, index: Any, values: Tensor):
        """Set or modify data values at the given indices to `values`.

        The default implementation edits the `buffer` field of this :class:`Data` object.
        Users may wish to override it in cases where the buffer is not used directly.

        Parameters
        ----------
        index: Any
            Indices to modify.

        values: Tensor
            Values to set data at indices to.

        """
        self.buffer[index] = values

    def normalize(self, p: int = 1, axis: int = 1) -> Tensor:
        """
        Apply Lp normalization along a specified axis to the buffer.

        Parameters
        ----------
        p: int, optional
            The order of the norm (default is 1, which corresponds to L1 normalization).

        axis: int, optional
            The axis along which to normalize (default is 0).

        Returns
        -------
        Tensor
            The normalized buffer.

        """
        self.buffer = torch.as_tensor(norm(self.buffer.numpy(), axis=axis, norm=f"l{p}"))
        return self.buffer

    def save(self):
        """Dump the buffer contents and metadata to disk at `root`.

        Since different datasets and pipelines call for different formats, implementation is left to the user.

        """
        pass

    def scan(self, pattern: str = "*", recursive: bool = False) -> list[str]:
        """Scans the `root` directory and returns a list of files found that match the glob `pattern`.

        Parameters
        ----------
        pattern: str, optional
            Pattern to look for in file names that should be included, glob-formatted. Defaults to "*" (any).

        recursive: bool, optional
            Whether glob should search recursively. Defaults to `False`.

        """
        files = []
        if os.path.exists(self.root):
            for f in glob(os.path.join(self.root, pattern), recursive=recursive):
                files.append(f)

        return natsorted(files)

    def sample(self, method: Callable, axis: int = 0, **kwargs):
        """Applies `method` to sample from this dataset once without mutating it, returning
        a copy of the object containing only the data and labels at the sampled indices.

        The method can be a reference to a :class:`~sklearn.model_selection.BaseCrossValidator`.
        In that case, the keyword arguments should include any applicable keyword arguments,
        e.g. `shuffle`, `label_key`, `group_key` if applicable (see also :class:`~data.sampling.CV`).

        If `method` is not a base cross validator, keyword arguments will be passed to it directly.

        Parameters
        ----------
        method: Callable or BaseCrossValidator
            Method used to sample from this dataset.

        axis: int, optional
            Axis along which selection is performed. Defaults to zero (that is, rows/buffer).

        kwargs:
            retain: int or float
                The number or proportion of buffer to retain.

            shuffle: bool
                If using a :class:`sklearn.model_selection.BaseCrossValidator` to sample,
                whether to toggle the `shuffle` parameter on.

            label_keys: str or list of str
                Label key(s) by which to stratify sampling, if applicable.

            group_key: str or list of str
                Label key by which to group sampling, if applicable.

        Returns
        -------
        Data
            A subset of this dataset.

        """
        # determine whether a BaseCrossValidator was provided and if so, set it up.
        # `method` must be called because sklearn CV objects are ABCMeta (hence unrecognizable) until called.
        try:
            is_cv = isinstance(method(), BaseCrossValidator)
        except TypeError:
            is_cv = False

        if is_cv:
            # determine the number of folds based on `retain` and the dataset length as defined by __len__.
            # if `retain` not provided as a keyword argument, treat this as a 2-split.
            retain = kwargs.get("retain", 0.5)
            folds = int(self.__len__() * retain) if isinstance(retain, float) else self.__len__() // retain

            # if `shuffle` provided as a keyword argument, use its value in the method call (default to True).
            validator = method(folds, shuffle=kwargs.get("shuffle", True))
            cv_iterator = CV(data=self, cross_validator=validator, **kwargs)

            # use the first fold.
            _, subset = next(iter(cv_iterator))

        elif isinstance(method, BalancedSampler):
            # special case of the balanced sampler, which uses the aggregated descriptor dataframe.
            subset = method(frame=self.metadata.to_dataframe(axis), **kwargs)

        else:
            # in all other cases, directly apply the method provided.
            subset = method(**kwargs)

        # trim buffer and labels, returning a new partial dataset without mutating the original.
        return self.trim(index=subset, axis=axis)

    def trim(self, index: Any, axis: int = None):
        """Trims this instance by selecting `indices`, potentially along `axis`, returning a subset of the original
        dataset in terms of both buffer entries and labels/descriptors. Does not mutate the underlying object.

        Parameters
        ----------
        index: Any
            Index(es) to retain.

        axis: int, optional
            Single axis to apply index selection to. Defaults to None, in which case `index` is used directly
            within :meth:`access`.

        """
        # create a deep copy of this data object and trim its buffer and metadata.
        partial_data = deepcopy(self)
        partial_data.buffer = self.access(index, axis=axis)

        for k in partial_data.get_metadata():
            if partial_data.metadata[k].axis == axis or axis is None:
                partial_data.metadata[k] = partial_data.metadata[k][index]

        return partial_data

    def ingest(self, folds: int, shots: int, keys: str | Sequence[str], replace: bool = False):
        """Loads and samples the data, returning a balanced subset thereof.

        Parameters
        ----------
        folds: int
            Number of cross validation folds.

        shots: int
            Number of samples per class.

        keys: Sequence[str]
            Keys by which to group the data.

        replace: bool, optional
            Whether to sample with replacement. Defaults to `False`.

        """
        raise NotImplementedError



import pytest
import os

from sapicore.data.external.drift import DriftDataset
from sapicore.tests import ROOT

TEST_ROOT = os.path.join(ROOT, "tests", "data", "test_data")  # test root directory.


class TestDrift:
    @pytest.mark.unit
    def test_drift_processing(self):
        # calling a dataset invokes its load() method and returns a self reference.
        try:
            DriftDataset(root=os.path.join(TEST_ROOT, "drift")).load()

        except ConnectionError:
            # package tests should NOT fail when a particular hardcoded archive server is down.
            pytest.mark.skip("Could not download drift dataset")


if __name__ == "__main__":
    pytest.main(args=["-s", "-v"])



""" Writes scalar and image data to tensorboard. """
import os

import numpy as np
import h5py as hdf

import torch
from torch.utils.tensorboard import SummaryWriter

import matplotlib.pyplot as plt
from sapicore.utils.plotting import spike_raster


__all__ = ("HDFData", "TensorboardWriter")


class HDFData:
    """Reads HDF5 data logged to disk by :class:`~utils.io.DataAccumulatorHook`.

    Parameters
    ----------
    path: str, optional
        Path to HDF5 data file.

    key: str, optional
        Specific key to read from HDF5. If not provided, the entire file will be read.

    buffer: dict, optional
        For initializing an HDFData object with a preexisting dictionary, e.g. to save hyperparameters
        computed at runtime.

    Warning
    -------
    This class assumes that the HDF5 file only contains labeled datasets, not groups, in accordance with
    the current implementation of :class:`~utils.io.DataAccumulatorHook`.
    A generalized data loading protocol is being developed.

    """

    def __init__(self, path: str = None, key: str = None, buffer: dict = None):
        self.path = path
        self.key = key

        # buffer into which data stored in `keys` on files found in `path` will be loaded on request.
        self.buffer = buffer if buffer is not None else {}

        if self.path:
            self.read()

    def read(self):
        """Extracts desired datasets and/or fields from HDF5 file and adds them to the buffer."""
        #
        with hdf.File(os.path.realpath(self.path), "r") as f:
            # add component identifier and class for reference.
            try:
                self.buffer["identifier"] = f.attrs["identifier"]
                self.buffer["class"] = f.attrs["class"]

            except (AttributeError, ValueError):
                self.buffer["identifier"] = "NA"
                self.buffer["class"] = "NA"

            # add configurable property data to buffer.
            for attr in list(f.attrs):
                if not self.key or (self.key == attr):
                    self.buffer[attr] = f.attrs[attr]

            # add loggable property data to buffer.
            for attr in list(f):
                if not self.key or (self.key == attr):
                    self.buffer[attr] = f[attr][:]


class TensorboardWriter:
    """Reusable SummaryWriter instance endowed with methods handling tensorboard file I/O given a data object.

    For ease of use, :meth:`write` can be called with a multi-data buffer, e.g. in cases where
    the same plotting settings apply to all objects.

    Parameters
    ----------
    log_dir: str
        Path to target tensorboard directory.

    """

    def __init__(self, log_dir: str):
        """Initializes a `SummaryWriter` object."""
        self.writer = SummaryWriter(log_dir=log_dir)

    def write(self, data: HDFData, key: str, **settings):
        """Writes data to tensorboard using `settings`.

        Parameters
        ----------
        data: HDFData
            Object holding buffered data as a dictionary.

        key: str
            Key of configurable or loggable to process in this iteration.

        settings:
            The following optional arguments control how data is visualized:

            kind: str
                Kind of plot or data to write, currently supports "raster", "trace", "heatmap", "hparam".

            format: str
                Whether to write a 2D time series as an "image" or a "scalar" data to tensorboard.
                Will default to image for 3D time series (e.g., weight matrix + time).

            size: tuple of int
                Two-item tuple specifying figure size to be passed on to plt.rcParams["figure.figsize"].

            step: int
                Where applicable, size of skip step when plotting a time series on an interactive slider plot.
                E.g., we don't want to plot a weight matrix for every 1.0 ms simulation step, so we set step=10.

            time_axis: int
                Specify which axis of the array corresponds to time steps. Used for stacking images in
                interactive tensorboard slider plots.

        """
        # only process keys that exist in the file being handled.
        if key in data.buffer.keys():
            # extract component identifier from data buffer.
            name = data.buffer.get("identifier")

            if settings.get("kind") == "raster":
                # handle the raster plot case.
                self.raster(name=name, array=data.buffer[key], settings=settings)

            elif settings.get("kind") == "trace":
                # handle the individual trace case as either an image or a scalar.
                self.trace(name=name, attr=key, array=data.buffer[key], settings=settings)

            elif settings.get("kind") == "heatmap":
                # handle the heatmap case, use interactive slider plots if provided multiple.
                self.heatmap(name=name, attr=key, array=data.buffer[key], settings=settings)

            elif settings.get("kind") == "hparam":
                self.hparam(name=name, attr=key, array=data.buffer[key], settings=settings)

    def heatmap(self, name: str, attr: str, array: np.ndarray, settings: dict):
        """Adds 2D numeric data to tensorboard as a heatmap."""
        # handle case where user is trying to save a single heatmap as opposed to a time series thereof.
        if len(array.shape) != 3:
            # expand array along what would be the default time axis (rows).
            array = array[None, :]

        # extract number of steps from time_axis (or default to row dimension).
        num_steps = array.shape[settings.get("time_axis", 0)]

        # set an appropriate figure size (heatmaps are always images).
        plt.rcParams["figure.figsize"] = settings.get("size", (15, 10))
        plt.figure()

        for j in range(0, num_steps, settings.get("step", 100)):
            plt.grid(False)

            time_slice = array.take(indices=j, axis=settings.get("time_axis", 0))
            p = plt.imshow(time_slice, cmap=plt.cm.viridis, aspect="auto")

            limit = np.max(np.abs(array))
            plt.clim(-limit, limit)

            plt.colorbar()
            self.writer.add_figure(f"{name}/{attr}", p.get_figure(), global_step=j)

    def raster(self, name: str, array: np.ndarray | torch.Tensor, settings: dict):
        """Adds 1D integer binary data to tensorboard as a raster plot."""
        # set an appropriate figure size.
        plt.rcParams["figure.figsize"] = settings.get("size", (15, 10))

        # add a raster plot for the quantity (always an image).
        fig = plt.figure()
        ax = fig.add_subplot()

        collections = spike_raster(
            data=array,
            line_size=settings.get("line_size", 0.25),
            events=settings.get("events"),
            event_colors=settings.get("event_colors"),
        )

        if collections is not None:
            _ = [ax.add_collection(collection) for collection in collections]
            self.writer.add_figure(tag=f"{name}/raster", figure=fig)

    def trace(self, name: str, attr: str, array: np.ndarray, settings: dict):
        """Adds 1D numeric trace data to tensorboard as a scalar plot or an image."""
        time_axis = settings.get("time_axis", 0)
        data_axis = 1 if time_axis == 0 else 0

        for j in range(array.shape[data_axis]):
            if settings.get("format") == "scalar":
                # handle the scalar plot case
                for i in range(array.shape[time_axis]):
                    self.writer.add_scalars(f"{name}/N{j}", {f"{attr}": array[i, j]}, global_step=i)

            else:
                # set figure size and initialize empty figure.
                plt.rcParams["figure.figsize"] = settings.get("size", (10, 5))

                # generate a line plot for our scalar.
                fig = plt.figure()
                plt.plot(array[:, j])

                # set y-axis limits slightly beyond min and max values in data.
                height = (np.max(array) - np.min(array)) * 1.1
                if height > 0:
                    plt.ylim(np.mean(array) - height, np.mean(array) + height)

                # add figure to tensorboard.
                self.writer.add_figure(tag=f"{name}/N{j}/{attr}", figure=fig)

    def hparam(self, name: str, attr: str, array: np.ndarray, settings: dict):
        """Writes singleton hyperparameters, such as configurable attributes, to tensorboard.

        If you would like to visualize non-singleton configurable attributes (e.g., synaptic delays), please
        use one of the other plotting methods.

        Warning
        -------
        Tensorboard's `add_hparams()` will not allow writing different hparams sequentially (after the first one, they
        turn into scalars). It also cannot handle writing of a hyperparameter without an accompanying metric.

        Any workaround will likely require patching tensorboard, a task left to the intrepid maintainer.
        For now, configurable attributes can be logged as scalars for easy viewing of parameter value distributions.

        Sapicore may end up using a better maintained data visualization library.

        """
        # in the scalar case, iterate over elements being described (e.g., neurons in an ensemble), necessarily 1D.
        if settings.get("format") == "scalar":
            for i in range(array.shape[0]):
                self.writer.add_scalars(f"{name}/N{i}", {f"{attr}": array[i]})



""" File and user I/O. """
from typing import Callable, Sequence

import _io

import os
import sys
import logging

from pathlib import Path
from itertools import chain

import yaml
import h5py as hdf

import torch
from torch import Tensor
from torch.nn import Module

from alive_progress import alive_bar
from tree_config import apply_config, load_config

from sapicore import __version__
from sapicore.utils.tensorboard import TensorboardWriter, HDFData

__all__ = (
    "DataAccumulatorHook",
    "MonitorHook",
    "ensure_dir",
    "flatten",
    "load_yaml",
    "save_yaml",
    "log_settings",
    "load_apply_config",
    "plot_tensorboard",
)

CHUNK_SIZE = 50.0
""" Data chunks will be dumped to disk after reaching this size in MB. """


class MonitorHook(Module):
    """Accumulates selected attribute data in memory on every forward call. Does not write to disk.

    Parameters
    ----------
    component: torch.nn.Module
        Network component whose attributes should be buffered in memory.

    attributes: list of str
        Instance attributes to be logged. Defaults to the full list `component.loggable_props`.

    entries: int
        Expected number of simulation steps. Used to preallocate the buffer.
        If unknown, `torch.vstack()` is used, which may result in slower performance.

    dtype:
        Data type.

    """

    def __init__(self, component: torch.nn.Module, attributes: Sequence[str], entries: int = None, dtype=torch.float):
        super().__init__()

        self.component = component
        self.attributes = component.loggable_props if not attributes else attributes

        self.iteration = 0
        self.entries = entries

        self.cache = {}
        self.dtype = dtype

        # (de)activate hook without removing it.
        self.active = True
        self.handle = component.register_forward_hook(self.monitor_hook())

    def __getitem__(self, item):
        if item is not None and not isinstance(item, slice):
            return self.cache[item] if item in self.cache.keys() else {}
        else:
            # [:] will be a shorthand for returning the cache dictionary.
            return self.cache

    def remove(self):
        self.handle.remove()

    def pause(self):
        self.active = False

    def unpause(self):
        self.active = True

    def monitor_hook(self) -> Callable:
        def fn(_, __, output):
            if self.active:
                # add current outputs to this data accumulator instance's cache dictionary, whose values are tensors.
                for attr in self.attributes:
                    odim = output[attr].dim()
                    if attr not in self.cache.keys():
                        # the cache dictionary is empty because this is the first iteration.
                        if self.entries is None:
                            # expand output by one dimension (zero axis) to fit.
                            self.cache[attr] = output[attr][None, :] if odim == 1 else output[attr][None, :, :]
                        else:
                            # preallocate if number of steps is known.
                            dim = [self.entries, len(output[attr])] + ([output[attr].shape[1]] if odim == 2 else [])
                            self.cache[attr] = torch.empty(dim, device=self.component.device, dtype=self.dtype)

                    else:
                        if self.entries is None:
                            # vertically stack output attribute to cache tensor at the appropriate key.
                            target = output[attr][None, :] if odim == 1 else output[attr][None, :, :]
                            self.cache[attr] = torch.vstack([self.cache[attr], target])
                        else:
                            # update appropriate row in preallocated tensor.
                            self.cache[attr][self.iteration] = output[attr]

                # advance iteration counter.
                self.iteration += 1

        return fn


class DataAccumulatorHook(Module):
    """Wraps any :class:`~torch.nn.Module` object, accumulating selected attribute data on every forward call.

    If used with a :class:`~engine.component.Component`, selectable attributes are those included in the dictionary
    returned by the Module's forward method, and are a subset of the object's `_loggable_props_`.

    Parameters
    ----------
    component: torch.nn.Module
        Reference to the network component object whose loggable buffered attributes should be dumped to disk.

    log_dir: str
        Path to the desired data loggable directory.

    attributes: list of str
        Instance attribute names to be logged. Defaults to the full list `component.loggable_props`.

    entries: int
        Number of log entries, equal to the number of simulation steps. Used to ensure the last batch in the run
        is written to disk despite its size being smaller than :attr:`CHUNK_SIZE`.

    """

    def __init__(self, component: torch.nn.Module, log_dir: str, attributes: Sequence[str], entries: int):
        super().__init__()

        self.log_dir = log_dir
        self.component = component
        self.attributes = component.loggable_props if not attributes else attributes
        self.hdf_file_path = os.path.join(self.log_dir, self.component.identifier + ".h5")

        # total number of forward calls to expect in this simulation run (steps).
        self.entries = entries

        # tensor for caching input until memory threshold is reached, at which point a chunk will be saved to disk.
        self.cache = {}
        self.chunk_size = {}

        # counter for iterations, to be checked against `entries` to ensure all data is written to disk.
        self.iterations = 0

        # initialize HDF5 files and register the forward hook.
        self.initialize_hdf()
        component.register_forward_hook(self.save_outputs_hook())

    def initialize_hdf(self) -> None:
        """Initializes HDF5 file on the first iteration of this data accumulator instance."""
        # detects whether we are in the first iteration before creating or appending to the file.
        if not os.path.exists(self.hdf_file_path):
            ensure_dir(os.path.dirname(self.hdf_file_path))
            with hdf.File(os.path.join(self.log_dir, self.component.identifier + ".h5"), "a") as hf:
                # write metadata to be able to identify this HDF with its object.
                hf.attrs["identifier"] = self.component.identifier
                hf.attrs["class"] = type(self.component).__name__
                hf.attrs["version"] = __version__

                # create datasets for loggable properties.
                # max shape unlimited in first axis, so that it can be expanded on every simulation step.
                for attr in self.attributes:
                    attr_tensor = getattr(self.component, attr)
                    expanded_dimensions = attr_tensor[None, :].shape
                    self.chunk_size[attr] = 0

                    hf.create_dataset(
                        name=attr,
                        shape=expanded_dimensions,
                        compression="lzf",
                        chunks=True,
                        maxshape=tuple([s if j > 0 else None for j, s in enumerate(expanded_dimensions)]),
                    )

                # write configurable properties ONCE for future reference, e.g. during tensorboard loggable.
                if hasattr(self.component, "_config_props"):
                    hf.create_group(name="configuration")
                    for prop in getattr(self.component, "_config_props"):
                        # log configurable attribute values as metadata in the "configuration" group.
                        value = getattr(self.component, prop)
                        hf["/configuration"].attrs[prop] = repr(value)
        else:
            raise FileExistsError("Target HDF5 log file already exists.")

    def save_outputs_hook(self) -> Callable:
        def fn(_, __, output):
            """Updates the output cache and appends to destination HDF5 file if cache is full or simulation ends."""
            # increment iteration counter.
            self.iterations += 1

            # add current outputs to this data accumulator instance's cache dictionary, whose values are tensors.
            for attr in self.attributes:
                if attr not in self.cache.keys():
                    # the cache dictionary is empty because this is the first iteration post-dump or in general.
                    self.cache[attr] = output[attr][None, :]  # expand output by one dimension (zero axis) to fit.
                else:
                    # vertically stack output attribute to cache tensor at the appropriate key.
                    self.cache[attr] = torch.vstack([self.cache[attr], output[attr][None, :]])

            # calculate cache size (MB).
            cache_size = sum([v.element_size() * v.nelement() for _, v in self.cache.items()]) / 1000000.0

            # if memory usage exceeds the limit given or reached the end of the run, append cached content to HDF5.
            if cache_size > CHUNK_SIZE or self.iterations >= self.entries:
                with hdf.File(os.path.join(self.log_dir, self.component.identifier + ".h5"), "a") as hf:
                    for attr in self.attributes:
                        # increase data chunk size by number of rows in the cache tensor.
                        self.chunk_size[attr] += self.cache[attr].shape[0]
                        hf[attr].resize(self.chunk_size[attr], axis=0)

                        # write current cache tensor to its respective HDF5 data structure at the correct indices.
                        start_index = hf[attr].shape[0] - self.cache[attr].shape[0]
                        end_index = hf[attr].shape[0]

                        # detach().numpy() necessary if the tensor requires grad.
                        hf[attr][start_index:end_index, :] = self.cache[attr].cpu().detach().numpy()

                    # clear cache.
                    self.cache = {}

        return fn

    def forward(self, data: Tensor) -> None:
        _ = self.component(data)
        return


def log_settings(
    level: int = logging.INFO,
    stream: _io.TextIOWrapper = sys.stdout,
    file: str = None,
    formatting: str = "%(asctime)s [%(levelname)s] %(message)s",
):
    """Temporary basic logger configuration.

    Parameters
    ----------
    level: int
        Logger level, `INFO` by default.

    stream:
        Destination stream, `stdout` by default.

    file:
        Destination file, `None` by default.

    formatting: str
        Message formatting.

    """
    logging.basicConfig(
        level=level,
        format=formatting,
        handlers=[logging.StreamHandler(stream)] + ([logging.FileHandler(file)] if file else []),
    )


def load_apply_config(path: str, apply_to: object = None) -> dict:
    # parse configuration from YAML if given and use resulting dictionary to initialize attributes.
    content = load_config(None, path) if os.path.exists(path) else None

    if not content:
        raise FileNotFoundError(f"Could not find a configuration YAML at {os.path.realpath(path)}")

    else:
        # apply the configuration to this pipeline object.
        configuration = content

        if apply_to:
            apply_config(apply_to, configuration)

    return configuration


def ensure_dir(path: str = None) -> str:
    """Ensures that a path is created if parts thereof do not exist yet and returns it.

    Parameters
    ----------
    path: str
        File path to be verified or created.

    Returns
    -------
    path: str
        The same file path, so it can be used in the calling context (e.g., if :func:`ensure_dir`
        wraps an assignment).

    """
    os.makedirs(path, exist_ok=True)
    return path


def flatten(lst: list[list]) -> list:
    """Flattens a list."""
    return list(chain(*lst))


def load_yaml(path: str) -> dict:
    """Parses the YAML at `path`, returning a dictionary."""
    return yaml.safe_load(Path(path).read_text())


def save_yaml(data: dict, path: str):
    """Saves a dictionary `data` to the YAML file specified by `path`."""
    with open(path, "w") as f:
        yaml.dump(data, f, default_flow_style=False)


def plot_tensorboard(configuration: dict, data_dir: str, tb_dir: str):
    """Loads simulation output data from disk and writes it to tensorboard for visual inspection."""
    # REFACTOR when rewriting this more formally, need a way to pass arbitrary plot settings that can't
    #  necessarily go through YAML, e.g. experiment event timestamps and label colors.
    writer = TensorboardWriter(log_dir=tb_dir)
    with alive_bar(total=len(os.listdir(data_dir)), force_tty=True) as bar:
        logging.info("Writing tensorboard data for visualization.")
        for file in os.listdir(data_dir):
            if not file.endswith(".h5"):
                continue

            # write what was asked.
            tensorboard_settings = configuration.get("simulation", {}).get("tensorboard", 0)
            if isinstance(tensorboard_settings, list):
                for task in tensorboard_settings:
                    for attr, settings in task.items():
                        # read only requested key from file.
                        data = HDFData(path=os.path.join(data_dir, file), key=attr)

                        # write the loggable or configurable attribute corresponding to this task.
                        writer.write(data, attr, **settings)
            bar()



""" Unit, integration, and functional tests. """
import os
import torch

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__), ".."))
""" Project root directory (.../sapicore), for use by test suite scripts. """

TEST_DEVICE = torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")
""" Hardware device to use during testing. Defaults to GPU when available, CPU otherwise. """



"""Integrator objects provide numerical methods for approximating ODE solutions."""
from torch import Tensor

from typing import Callable
from sapicore.utils.constants import DT

__all__ = ("Integrator", "RungeKutta")


class Integrator:
    """Base integrator class.

    Parameters
    ----------
    identifier: str, optional
        Name of the integrator.

    step: float, optional
        Step size. Defaults to :attr:`~utils.constants.DT`.

    Warning
    -------
    Users are encouraged to use the :class:`RungeKutta` object in their :meth:`~engine.neuron.Neuron.integrate`
    implementations (which in turn should be used in :meth:`~engine.neuron.Neuron.forward`), regardless of whether
    they anticipate using more advanced approximation methods.

    """

    def __init__(self, identifier: str = None, step: int = DT, **kwargs):
        self.identifier = identifier
        self.step = step

    def __call__(self, x: Tensor, equation: Callable, **kwargs) -> Tensor:
        """Approximates the next value of the ODE `equation`, whose keyword arguments are given as `kwargs`,
        for an initial value `x` and a time step `step` (defaults to the global :attr:`~utils.constants.DT`).

        Returns
        -------
        Tensor
            Approximate value(s) of the function at t+DT, where t is the time `x` was obtained (current time).

        Raises
        ------
        NotImplementedError
            The __call__ method must be implemented by each derived :class:`~utils.integration.Integrator`.

        """
        raise NotImplementedError


class RungeKutta(Integrator):
    """Generic Runge-Kutta integrator.

    When called, approximates the next value of an ODE `equation`, whose keyword arguments are given as `kwargs`,
    for an initial value `x` and a time step `step` (defaults to the global :attr:`~utils.constants.DT`),
    using the Runge-Kutta method of order `order`.

    Parameters
    ----------
    identifier: str, optional
        Name of the integrator.

    order: int, optional
        Order of the integrator. RK1 and RK4 are currently supported.

    Note
    ----
    The forward Euler method is Runge-Kutta of order 1.

    References
    ----------
    `Tutorial <https://perso.crans.org/besson/publis/notebooks/Runge-Kutta_methods_for_ODE_integration_in_Python.html>`_
        :class:`~utils.integration.RungeKutta` implements the algorithm variants described above.

    """

    def __init__(self, identifier: str = "RK", order: int = 1, **kwargs):
        super().__init__(identifier=identifier, **kwargs)
        self.order = order

    def __call__(self, x: Tensor, equation: Callable, **kwargs) -> Tensor:
        """Uses the update rule `equation` to compute a Runge-Kutta approximation at t+DT.
        Variables needed by the difference equation are passed on by the caller as keyword arguments.

        Example
        -------
        Initialize LIF neuron specifying RK4 as the integrator:

            >>> from sapicore.engine.ensemble.spiking import LIFEnsemble
            >>> layer = LIFEnsemble(num_units=10, integrator=RungeKutta(order=4))

        Peek at the RK4 approximated value for `voltage` at t+DT (this will NOT modify the voltage value in-place):

            >>> data = Tensor(list(range(10)))
            >>> layer.integrate(data=data)

        The default LIF forward method makes use of the Integrator interface to update the voltage:

            >>> layer(data)
            >>> print(layer.voltage)

        """
        if self.order == 1:
            # standard forward Euler approximation.
            return x + equation(x=x, **kwargs) * self.step

        else:
            # compute RK4 slope estimates and return their weighted average.
            k1 = equation(x=x, **kwargs)
            k2 = equation(x=x + (self.step / 2.0) * k1, **kwargs)
            k3 = equation(x=x + (self.step / 2.0) * k2, **kwargs)
            k4 = equation(x=x + self.step * k3, **kwargs)

            return x + (self.step / 6.0) * (k1 + 2.0 * k2 + 2.0 * k3 + k4)
